# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from .i18n import _
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .settings import (
				KIND as _KIND,
				Setting as _Setting,
				RegisterRW as _RegisterRW,
				FeatureRW as _FeatureRW,
				BooleanValidator as _BooleanV,
				ChoicesValidator as _ChoicesV,
			)

_DK = _hidpp10.DEVICE_KIND
_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
# pre-defined basic setting descriptors
#

def register_toggle(name, register,
					true_value=_BooleanV.default_true,
					false_value=_BooleanV.default_false,
					mask=_BooleanV.default_mask,
					label=None, description=None, device_kind=None):
	validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
	rw = _RegisterRW(register)
	return _Setting(name, rw, validator, label=label, description=description, device_kind=device_kind)


def register_choices(name, register, choices,
					kind=_KIND.choice,
					label=None, description=None, device_kind=None):
	assert choices
	validator = _ChoicesV(choices)
	rw = _RegisterRW(register)
	return _Setting(name, rw, validator, kind=kind, label=label, description=description, device_kind=device_kind)


def feature_toggle(name, feature,
					read_function_id=_FeatureRW.default_read_fnid,
					write_function_id=_FeatureRW.default_write_fnid,
					true_value=_BooleanV.default_true,
					false_value=_BooleanV.default_false,
					mask=_BooleanV.default_mask,
					label=None, description=None, device_kind=None):
	validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
	rw = _FeatureRW(feature, read_function_id, write_function_id)
	return _Setting(name, rw, validator, label=label, description=description, device_kind=device_kind)

#
# common strings for settings
#

_SMOOTH_SCROLL = ('smooth-scroll', _("Smooth Scrolling"),
							_("High-sensitivity mode for vertical scroll with the wheel."))
_SIDE_SCROLL = ('side-scroll', _("Side Scrolling"),
							_("When disabled, pushing the wheel sideways sends custom button events\n"
							"instead of the standard side-scrolling events."))
_DPI = ('dpi', _("Sensitivity (DPI)"), None)
_FN_SWAP = ('fn-swap', _("Swap Fx function"),
							_("When set, the F1..F12 keys will activate their special function,\n"
						 	"and you must hold the FN key to activate their standard function.")
						 	+ '\n\n' +
						 	_("When unset, the F1..F12 keys will activate their standard function,\n"
						 	"and you must hold the FN key to activate their special function."))
_HAND_DETECTION = ('hand-detection', _("Hand Detection"),
							_("Turn on illumination when the hands hover over the keyboard."))

#
#
#

def _register_hand_detection(register=_R.keyboard_hand_detection,
					true_value=b'\x00\x00\x00', false_value=b'\x00\x00\x30', mask=b'\x00\x00\xFF'):
	return register_toggle(_HAND_DETECTION[0], register, true_value=true_value, false_value=false_value,
					label=_HAND_DETECTION[1], description=_HAND_DETECTION[2],
					device_kind=_DK.keyboard)

def _register_fn_swap(register=_R.keyboard_fn_swap, true_value=b'\x00\x01', mask=b'\x00\x01'):
	return register_toggle(_FN_SWAP[0], register, true_value=true_value, mask=mask,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=_DK.keyboard)

def _register_smooth_scroll(register=_R.mouse_button_flags, true_value=0x40, mask=0x40):
	return register_toggle(_SMOOTH_SCROLL[0], register, true_value=true_value, mask=mask,
					label=_SMOOTH_SCROLL[1], description=_SMOOTH_SCROLL[2],
					device_kind=_DK.mouse)

def _register_side_scroll(register=_R.mouse_button_flags, true_value=0x02, mask=0x02):
	return register_toggle(_SIDE_SCROLL[0], register, true_value=true_value, mask=mask,
 					label=_SIDE_SCROLL[1], description=_SIDE_SCROLL[2],
					device_kind=_DK.mouse)

def _register_dpi(register=_R.mouse_dpi, choices=None):
	return register_choices(_DPI[0], register, choices,
					label=_DPI[1], description=_DPI[2],
					device_kind=_DK.mouse)


def _feature_fn_swap():
	return feature_toggle(_FN_SWAP[0], _F.FN_INVERSION,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=_DK.keyboard)

def _feature_new_fn_swap():
	return feature_toggle(_FN_SWAP[0], _F.NEW_FN_INVERSION,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=_DK.keyboard)

def _feature_smooth_scroll():
	return feature_toggle(_SMOOTH_SCROLL[0], _F.HI_RES_SCROLLING,
					label=_SMOOTH_SCROLL[1], description=_SMOOTH_SCROLL[2],
					device_kind=_DK.mouse)

#
#
#

from collections import namedtuple
_SETTINGS_LIST = namedtuple('_SETTINGS_LIST', [
					'fn_swap',
					'new_fn_swap',
					'smooth_scroll',
					'side_scroll',
					'dpi',
					'hand_detection',
					'typing_illumination',
					])
del namedtuple

RegisterSettings = _SETTINGS_LIST(
				fn_swap=_register_fn_swap,
				new_fn_swap=None,
				smooth_scroll=_register_smooth_scroll,
				side_scroll=_register_side_scroll,
				dpi=_register_dpi,
				hand_detection=_register_hand_detection,
				typing_illumination=None,
			)
FeatureSettings =  _SETTINGS_LIST(
				fn_swap=_feature_fn_swap,
				new_fn_swap=_feature_new_fn_swap,
				smooth_scroll=_feature_smooth_scroll,
				side_scroll=None,
				dpi=None,
				hand_detection=None,
				typing_illumination=None,
			)

del _SETTINGS_LIST

#
#
#

def check_feature_settings(device, already_known):
	"""Try to auto-detect device settings by the HID++ 2.0 features they have."""
	if device.features is None:
		return
	if device.protocol and device.protocol < 2.0:
		return
	if not any(s.name == _FN_SWAP[0] for s in already_known) and _F.FN_INVERSION in device.features:
		fn_swap = FeatureSettings.fn_swap()
		already_known.append(fn_swap(device))
	if not any(s.name == _SMOOTH_SCROLL[0] for s in already_known) and _F.HI_RES_SCROLLING in device.features:
		smooth_scroll = FeatureSettings.smooth_scroll()
		already_known.append(smooth_scroll(device))



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from gi.repository import Gtk

from solaar import __version__, NAME
from solaar.i18n import _

#
#
#

_dialog = None


def _create():
	about = Gtk.AboutDialog()

	about.set_program_name(NAME)
	about.set_version(__version__)
	about.set_comments(_("Shows status of devices connected\nthrough wireless Logitech receivers."))

	about.set_logo_icon_name(NAME.lower())

	about.set_copyright('© 2012-2013 Daniel Pavel')
	about.set_license_type(Gtk.License.GPL_2_0)

	about.set_authors(('Daniel Pavel http://github.com/pwr',))
	try:
		about.add_credit_section(_("GUI design"), ('Julien Gascard', 'Daniel Pavel'))
		about.add_credit_section(_("Testing"), (
						'Douglas Wagner',
						'Julien Gascard',
						'Peter Wu http://www.lekensteyn.nl/logitech-unifying.html',
						))
		about.add_credit_section(_("Logitech documentation"), (
						'Julien Danjou http://julien.danjou.info/blog/2012/logitech-unifying-upower',
						'Nestor Lopez Casado http://drive.google.com/folderview?id=0BxbRzx7vEV7eWmgwazJ3NUFfQ28',
						))
	except TypeError:
		# gtk3 < ~3.6.4 has incorrect gi bindings
		import logging
		logging.exception("failed to fully create the about dialog")
	except:
		# the Gtk3 version may be too old, and the function does not exist
		import logging
		logging.exception("failed to fully create the about dialog")

	about.set_translator_credits('\n'.join((
					'Papoteur, David Geiger, Damien Lallement (français)',
					'Adrian Piotrowicz (polski)',
					'Daniel Pavel (română)',
					'Daniel Zippert, Emelie Snecker (svensk)',
					)))

	about.set_website('http://pwr.github.io/Solaar/')
	about.set_website_label(NAME)

	about.connect('response', lambda x, y: x.hide())

	def _hide(dialog, event):
		dialog.hide()
		return True
	about.connect('delete-event', _hide)

	return about


def show_window(trigger=None):
	global _dialog
	if _dialog is None:
		_dialog = _create()
	_dialog.present()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Generic Human Interface Device API.

It is currently a partial pure-Python implementation of the native HID API
implemented by signal11 (https://github.com/signal11/hidapi), and requires
``pyudev``.
The docstrings are mostly copied from the hidapi API header, with changes where
necessary.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os as _os
import errno as _errno
from select import select as _select
from pyudev import Context as _Context, Monitor as _Monitor, Device as _Device


native_implementation = 'udev'


# the tuple object we'll expose when enumerating devices
from collections import namedtuple
DeviceInfo = namedtuple('DeviceInfo', [
				'path',
				'vendor_id',
				'product_id',
				'serial',
				'release',
				'manufacturer',
				'product',
				'interface',
				'driver',
				])
del namedtuple


#
# exposed API
# docstrings mostly copied from hidapi.h
#

def init():
	"""This function is a no-op, and exists only to match the native hidapi
	implementation.

	:returns: ``True``.
	"""
	return True


def exit():
	"""This function is a no-op, and exists only to match the native hidapi
	implementation.

	:returns: ``True``.
	"""
	return True


def _match(action, device, vendor_id=None, product_id=None, interface_number=None, hid_driver=None):
	usb_device = device.find_parent('usb', 'usb_device')
	# print ("* parent", action, device, "usb:", usb_device)
	if not usb_device:
		return

	vid = usb_device.get('ID_VENDOR_ID')
	pid = usb_device.get('ID_MODEL_ID')
	if not ((vendor_id is None or vendor_id == int(vid, 16)) and
			(product_id is None or product_id == int(pid, 16))):
		return

	if action == 'add':
		hid_device = device.find_parent('hid')
		if not hid_device:
			return
		hid_driver_name = hid_device.get('DRIVER')
		# print ("** found hid", action, device, "hid:", hid_device, hid_driver_name)
		if hid_driver:
			if isinstance(hid_driver, tuple):
				if hid_driver_name not in hid_driver:
					return
			elif hid_driver_name != hid_driver:
				return

		intf_device = device.find_parent('usb', 'usb_interface')
		# print ("*** usb interface", action, device, "usb_interface:", intf_device)
		if interface_number is None:
			usb_interface = None if intf_device is None else intf_device.attributes.asint('bInterfaceNumber')
		else:
			usb_interface = None if intf_device is None else intf_device.attributes.asint('bInterfaceNumber')
			if usb_interface is None or interface_number != usb_interface:
				return

		attrs = usb_device.attributes
		d_info = DeviceInfo(path=device.device_node,
							vendor_id=vid[-4:],
							product_id=pid[-4:],
							serial=hid_device.get('HID_UNIQ'),
							release=attrs.get('bcdDevice'),
							manufacturer=attrs.get('manufacturer'),
							product=attrs.get('product'),
							interface=usb_interface,
							driver=hid_driver_name)
		return d_info

	elif action == 'remove':
		# print (dict(device), dict(usb_device))

		d_info = DeviceInfo(path=device.device_node,
							vendor_id=vid[-4:],
							product_id=pid[-4:],
							serial=None,
							release=None,
							manufacturer=None,
							product=None,
							interface=None,
							driver=None)
		return d_info


def monitor_glib(callback, *device_filters):
	from gi.repository import GLib

	c = _Context()

	# already existing devices
	# for device in c.list_devices(subsystem='hidraw'):
	# 	# print (device, dict(device), dict(device.attributes))
	# 	for filter in device_filters:
	# 		d_info = _match('add', device, *filter)
	# 		if d_info:
	# 			GLib.idle_add(callback, 'add', d_info)
	# 			break

	m = _Monitor.from_netlink(c)
	m.filter_by(subsystem='hidraw')

	def _process_udev_event(monitor, condition, cb, filters):
		if condition == GLib.IO_IN:
			event = monitor.receive_device()
			if event:
				action, device = event
				# print ("***", action, device)
				if action == 'add':
					for filter in filters:
						d_info = _match(action, device, *filter)
						if d_info:
							GLib.idle_add(cb, action, d_info)
							break
				elif action == 'remove':
					# the GLib notification does _not_ match!
					pass
		return True

	try:
		# io_add_watch_full may not be available...
		GLib.io_add_watch_full(m, GLib.PRIORITY_LOW, GLib.IO_IN, _process_udev_event, callback, device_filters)
		# print ("did io_add_watch_full")
	except AttributeError:
		try:
			# and the priority parameter appeared later in the API
			GLib.io_add_watch(m, GLib.PRIORITY_LOW, GLib.IO_IN, _process_udev_event, callback, device_filters)
			# print ("did io_add_watch with priority")
		except:
			GLib.io_add_watch(m, GLib.IO_IN, _process_udev_event, callback, device_filters)
			# print ("did io_add_watch")

	m.start()


def enumerate(vendor_id=None, product_id=None, interface_number=None, hid_driver=None):
	"""Enumerate the HID Devices.

	List all the HID devices attached to the system, optionally filtering by
	vendor_id, product_id, and/or interface_number.

	:returns: a list of matching ``DeviceInfo`` tuples.
	"""
	for dev in _Context().list_devices(subsystem='hidraw'):
		dev_info = _match('add', dev, vendor_id, product_id, interface_number, hid_driver)
		if dev_info:
			yield dev_info


def open(vendor_id, product_id, serial=None):
	"""Open a HID device by its Vendor ID, Product ID and optional serial number.

	If no serial is provided, the first device with the specified IDs is opened.

	:returns: an opaque device handle, or ``None``.
	"""
	for device in enumerate(vendor_id, product_id):
		if serial is None or serial == device.serial:
			return open_path(device.path)


def open_path(device_path):
	"""Open a HID device by its path name.

	:param device_path: the path of a ``DeviceInfo`` tuple returned by
	enumerate().

	:returns: an opaque device handle, or ``None``.
	"""
	assert device_path
	assert device_path.startswith('/dev/hidraw')
	return _os.open(device_path, _os.O_RDWR | _os.O_SYNC)


def close(device_handle):
	"""Close a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	"""
	assert device_handle
	_os.close(device_handle)


def write(device_handle, data):
	"""Write an Output report to a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	:param data: the data bytes to send including the report number as the
	first byte.

	The first byte of data[] must contain the Report ID. For
	devices which only support a single report, this must be set
	to 0x0. The remaining bytes contain the report data. Since
	the Report ID is mandatory, calls to hid_write() will always
	contain one more byte than the report contains. For example,
	if a hid report is 16 bytes long, 17 bytes must be passed to
	hid_write(), the Report ID (or 0x0, for devices with a
	single report), followed by the report data (16 bytes). In
	this example, the length passed in would be 17.

	write() will send the data on the first OUT endpoint, if
	one exists. If it does not, it will send the data through
	the Control Endpoint (Endpoint 0).
	"""
	assert device_handle
	assert data
	assert isinstance(data, bytes), (repr(data), type(data))
	bytes_written = _os.write(device_handle, data)
	if bytes_written != len(data):
		raise IOError(_errno.EIO, 'written %d bytes out of expected %d' % (bytes_written, len(data)))


def read(device_handle, bytes_count, timeout_ms=-1):
	"""Read an Input report from a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	:param bytes_count: maximum number of bytes to read.
	:param timeout_ms: can be -1 (default) to wait for data indefinitely, 0 to
	read whatever is in the device's input buffer, or a positive integer to
	wait that many milliseconds.

	Input reports are returned to the host through the INTERRUPT IN endpoint.
	The first byte will contain the Report number if the device uses numbered
	reports.

	:returns: the data packet read, an empty bytes string if a timeout was
	reached, or None if there was an error while reading.
	"""
	assert device_handle
	timeout = None if timeout_ms < 0 else timeout_ms / 1000.0
	rlist, wlist, xlist = _select([device_handle], [], [device_handle], timeout)

	if xlist:
		assert xlist == [device_handle]
		raise IOError(_errno.EIO, 'exception on file descriptor %d' % device_handle)

	if rlist:
		assert rlist == [device_handle]
		data = _os.read(device_handle, bytes_count)
		assert data is not None
		assert isinstance(data, bytes), (repr(data), type(data))
		return data
	else:
		return b''


_DEVICE_STRINGS = {
			0: 'manufacturer',
			1: 'product',
			2: 'serial',
}


def get_manufacturer(device_handle):
	"""Get the Manufacturer String from a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	"""
	return get_indexed_string(device_handle, 0)


def get_product(device_handle):
	"""Get the Product String from a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	"""
	return get_indexed_string(device_handle, 1)


def get_serial(device_handle):
	"""Get the serial number from a HID device.

	:param device_handle: a device handle returned by open() or open_path().
	"""
	serial = get_indexed_string(device_handle, 2)
	if serial is not None:
		return ''.join(hex(ord(c)) for c in serial)


def get_indexed_string(device_handle, index):
	"""Get a string from a HID device, based on its string index.

	Note: currently not working in the ``hidraw`` native implementation.

	:param device_handle: a device handle returned by open() or open_path().
	:param index: the index of the string to get.
	"""
	if index not in _DEVICE_STRINGS:
		return None

	assert device_handle
	stat = _os.fstat(device_handle)
	dev = _Device.from_device_number(_Context(), 'char', stat.st_rdev)
	if dev:
		hid_dev = dev.find_parent('hid')
		if hid_dev:
			assert 'HID_ID' in hid_dev
			bus, _ignore, _ignore = hid_dev['HID_ID'].split(':')

			if bus == '0003':  # USB
				usb_dev = dev.find_parent('usb', 'usb_device')
				assert usb_dev
				key = _DEVICE_STRINGS[index]
				attrs = usb_dev.attributes
				if key in attrs:
					return attrs[key]

			elif bus == '0005':  # BLUETOOTH
				# TODO
				pass



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
from select import select as _select
import time
from binascii import hexlify, unhexlify
import hidapi as _hid

#
#
#

try:
	read_packet = raw_input
except NameError:
	# Python 3 equivalent of raw_input
	read_packet = input

interactive = os.isatty(0)
prompt = '?? Input: ' if interactive else ''
start_time = time.time()

strhex = lambda d: hexlify(d).decode('ascii').upper()
try:
	unicode
	# this is certanly Python 2
	is_string = lambda d: isinstance(d, unicode)
	# no easy way to distinguish between b'' and '' :(
						# or (isinstance(d, str) \
						# 	and not any((chr(k) in d for k in range(0x00, 0x1F))) \
						# 	and not any((chr(k) in d for k in range(0x80, 0xFF))) \
						# 	)
except:
	# this is certanly Python 3
	# In Py3, unicode and str are equal (the unicode object does not exist)
	is_string = lambda d: isinstance(d, str)

#
#
#

from threading import Lock
print_lock = Lock()
del Lock

def _print(marker, data, scroll=False):
	t = time.time() - start_time
	if is_string(data):
		s = marker + ' ' + data
	else:
		hexs = strhex(data)
		s = '%s (% 8.3f) [%s %s %s %s] %s' % (marker, t, hexs[0:2], hexs[2:4], hexs[4:8], hexs[8:], repr(data))

	with print_lock:
		# allow only one thread at a time to write to the console, otherwise
		# the output gets garbled, especially with ANSI codes.

		if interactive and scroll:
			# scroll the entire screen above the current line up by 1 line
			sys.stdout.write('\033[s'  # save cursor position
							'\033[S'  # scroll up
							'\033[A'   # cursor up
							'\033[L'    # insert 1 line
							'\033[G')   # move cursor to column 1
		sys.stdout.write(s)
		if interactive and scroll:
			# restore cursor position
			sys.stdout.write('\033[u')
		else:
			sys.stdout.write('\n')

		# flush stdout manually...
		# because trying to open stdin/out unbuffered programatically
		# works much too differently in Python 2/3
		sys.stdout.flush()


def _error(text, scroll=False):
	_print('!!', text, scroll)


def _continuous_read(handle, timeout=2000):
	while True:
		try:
			reply = _hid.read(handle, 128, timeout)
		except OSError as e:
			_error("Read failed, aborting: " + str(e), True)
			break
		assert reply is not None
		if reply:
			_print('>>', reply, True)


def _validate_input(line, hidpp=False):
	try:
		data = unhexlify(line.encode('ascii'))
	except Exception as e:
		_error("Invalid input: " + str(e))
		return None

	if hidpp:
		if len(data) < 4:
			_error("Invalid HID++ request: need at least 4 bytes")
			return None
		if data[:1] not in b'\x10\x11':
			_error("Invalid HID++ request: first byte must be 0x10 or 0x11")
			return None
		if data[1:2] not in b'\xFF\x01\x02\x03\x04\x05\x06':
			_error("Invalid HID++ request: second byte must be 0xFF or one of 0x01..0x06")
			return None
		if data[:1] == b'\x10':
			if len(data) > 7:
				_error("Invalid HID++ request: maximum length of a 0x10 request is 7 bytes")
				return None
			while len(data) < 7:
				data = (data + b'\x00' * 7)[:7]
		elif data[:1] == b'\x11':
			if len(data) > 20:
				_error("Invalid HID++ request: maximum length of a 0x11 request is 20 bytes")
				return None
			while len(data) < 20:
				data = (data + b'\x00' * 20)[:20]

	return data


def _open(args):
	device = args.device
	if args.hidpp and not device:
		for d in _hid.enumerate(vendor_id=0x046d):
			if d.driver == 'logitech-djreceiver':
				device = d.path
				break
		if not device:
			sys.exit("!! No HID++ receiver found.")
	if not device:
		sys.exit("!! Device path required.")

	print (".. Opening device", device)
	handle = _hid.open_path(device)
	if not handle:
		sys.exit("!! Failed to open %s, aborting." % device)

	print (".. Opened handle %r, vendor %r product %r serial %r." % (
					handle,
					_hid.get_manufacturer(handle),
					_hid.get_product(handle),
					_hid.get_serial(handle)))
	if args.hidpp:
		if _hid.get_manufacturer(handle) != b'Logitech':
			sys.exit("!! Only Logitech devices support the HID++ protocol.")
		print (".. HID++ validation enabled.")
	else:
		if (_hid.get_manufacturer(handle) == b'Logitech' and
			b'Receiver' in _hid.get_product(handle)):
			args.hidpp = True
			print (".. Logitech receiver detected, HID++ validation enabled.")

	return handle

#
#
#

def _parse_arguments():
	import argparse
	arg_parser = argparse.ArgumentParser()
	arg_parser.add_argument('--history', help="history file (default ~/.hidconsole-history)")
	arg_parser.add_argument('--hidpp', action='store_true', help="ensure input data is a valid HID++ request")
	arg_parser.add_argument('device', nargs='?', help="linux device to connect to (/dev/hidrawX); "
							"may be omitted if --hidpp is given, in which case it looks for the first Logitech receiver")
	return arg_parser.parse_args()


def main():
	args = _parse_arguments()
	handle = _open(args)

	if interactive:
		print (".. Press ^C/^D to exit, or type hex bytes to write to the device.")

		import readline
		if args.history is None:
			import os.path
			args.history = os.path.join(os.path.expanduser('~'), '.hidconsole-history')
		try:
			readline.read_history_file(args.history)
		except:
			# file may not exist yet
			pass

	try:
		from threading import Thread
		t = Thread(target=_continuous_read, args=(handle,))
		t.daemon = True
		t.start()

		if interactive:
			# move the cursor at the bottom of the screen
			sys.stdout.write('\033[300B')  # move cusor at most 300 lines down, don't scroll

		while t.is_alive():
			line = read_packet(prompt)
			line = line.strip().replace(' ', '')
			# print ("line", line)
			if not line:
				continue

			data = _validate_input(line, args.hidpp)
			if data is None:
				continue

			_print('<<', data)
			_hid.write(handle, data)
			# wait for some kind of reply
			if args.hidpp and not interactive:
				rlist, wlist, xlist = _select([handle], [], [], 1)
				if data[1:2] == b'\xFF':
					# the receiver will reply very fast, in a few milliseconds
					time.sleep(0.010)
				else:
					# the devices might reply quite slow
					time.sleep(0.700)
	except EOFError:
		if interactive:
			print ("")
		else:
			time.sleep(1)

	finally:
		print (".. Closing handle %r" % handle)
		_hid.close(handle)
		if interactive:
			readline.write_history_file(args.history)


if __name__ == '__main__':
	main()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Base low-level functions used by the API proper.
# Unlikely to be used directly unless you're expanding the API.

from __future__ import absolute_import, division, print_function, unicode_literals

from time import time as _timestamp
from random import getrandbits as _random_bits

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger


from .common import strhex as _strhex, KwException as _KwException, pack as _pack
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
import hidapi as _hid

#
#
#

_SHORT_MESSAGE_SIZE = 7
_LONG_MESSAGE_SIZE = 20
_MEDIUM_MESSAGE_SIZE = 15
_MAX_READ_SIZE = 32

"""Default timeout on read (in seconds)."""
DEFAULT_TIMEOUT = 4
# the receiver itself should reply very fast, within 500ms
_RECEIVER_REQUEST_TIMEOUT = 0.9
# devices may reply a lot slower, as the call has to go wireless to them and come back
_DEVICE_REQUEST_TIMEOUT = DEFAULT_TIMEOUT
# when pinging, be extra patient
_PING_TIMEOUT = DEFAULT_TIMEOUT * 2

#
# Exceptions that may be raised by this API.
#

class NoReceiver(_KwException):
	"""Raised when trying to talk through a previously open handle, when the
	receiver is no longer available. Should only happen if the receiver is
	physically disconnected from the machine, or its kernel driver module is
	unloaded."""
	pass


class NoSuchDevice(_KwException):
	"""Raised when trying to reach a device number not paired to the receiver."""
	pass


class DeviceUnreachable(_KwException):
	"""Raised when a request is made to an unreachable (turned off) device."""
	pass

#
#
#

from .base_usb import ALL as _RECEIVER_USB_IDS

def receivers():
	"""List all the Linux devices exposed by the UR attached to the machine."""
	for receiver_usb_id in _RECEIVER_USB_IDS:
		for d in _hid.enumerate(*receiver_usb_id):
			yield d


def notify_on_receivers_glib(callback):
	"""Watch for matching devices and notifies the callback on the GLib thread."""
	_hid.monitor_glib(callback, *_RECEIVER_USB_IDS)

#
#
#

def open_path(path):
	"""Checks if the given Linux device path points to the right UR device.

	:param path: the Linux device path.

	The UR physical device may expose multiple linux devices with the same
	interface, so we have to check for the right one. At this moment the only
	way to distinguish betheen them is to do a test ping on an invalid
	(attached) device number (i.e., 0), expecting a 'ping failed' reply.

	:returns: an open receiver handle if this is the right Linux device, or
	``None``.
	"""
	return _hid.open_path(path)


def open():
	"""Opens the first Logitech Unifying Receiver found attached to the machine.

	:returns: An open file handle for the found receiver, or ``None``.
	"""
	for rawdevice in receivers():
		handle = open_path(rawdevice.path)
		if handle:
			return handle


def close(handle):
	"""Closes a HID device handle."""
	if handle:
		try:
			if isinstance(handle, int):
				_hid.close(handle)
			else:
				handle.close()
			# _log.info("closed receiver handle %r", handle)
			return True
		except:
			# _log.exception("closing receiver handle %r", handle)
			pass

	return False


def write(handle, devnumber, data):
	"""Writes some data to the receiver, addressed to a certain device.

	:param handle: an open UR handle.
	:param devnumber: attached device number.
	:param data: data to send, up to 5 bytes.

	The first two (required) bytes of data must be the SubId and address.

	:raises NoReceiver: if the receiver is no longer available, i.e. has
	been physically removed from the machine, or the kernel driver has been
	unloaded. The handle will be closed automatically.
	"""
	# the data is padded to either 5 or 18 bytes
	assert data is not None
	assert isinstance(data, bytes), (repr(data), type(data))

	if len(data) > _SHORT_MESSAGE_SIZE - 2 or data[:1] == b'\x82':
		wdata = _pack('!BB18s', 0x11, devnumber, data)
	else:
		wdata = _pack('!BB5s', 0x10, devnumber, data)
	if _log.isEnabledFor(_DEBUG):
		_log.debug("(%s) <= w[%02X %02X %s %s]", handle, ord(wdata[:1]), devnumber, _strhex(wdata[2:4]), _strhex(wdata[4:]))

	try:
		_hid.write(int(handle), wdata)
	except Exception as reason:
		_log.error("write failed, assuming handle %r no longer available", handle)
		close(handle)
		raise NoReceiver(reason=reason)


def read(handle, timeout=DEFAULT_TIMEOUT):
	"""Read some data from the receiver. Usually called after a write (feature
	call), to get the reply.

	:param: handle open handle to the receiver
	:param: timeout how long to wait for a reply, in seconds

	:returns: a tuple of (devnumber, message data), or `None`

	:raises NoReceiver: if the receiver is no longer available, i.e. has
	been physically removed from the machine, or the kernel driver has been
	unloaded. The handle will be closed automatically.
	"""
	reply = _read(handle, timeout)
	if reply:
		return reply[1:]


def _read(handle, timeout):
	"""Read an incoming packet from the receiver.

	:returns: a tuple of (report_id, devnumber, data), or `None`.

	:raises NoReceiver: if the receiver is no longer available, i.e. has
	been physically removed from the machine, or the kernel driver has been
	unloaded. The handle will be closed automatically.
	"""
	try:
		# convert timeout to milliseconds, the hidapi expects it
		timeout = int(timeout * 1000)
		data = _hid.read(int(handle), _MAX_READ_SIZE, timeout)
	except Exception as reason:
		_log.error("read failed, assuming handle %r no longer available", handle)
		close(handle)
		raise NoReceiver(reason=reason)

	if data:
		assert isinstance(data, bytes), (repr(data), type(data))
		report_id = ord(data[:1])
		assert ((report_id & 0xF0 == 0) or
				(report_id == 0x10 and len(data) == _SHORT_MESSAGE_SIZE) or
				(report_id == 0x11 and len(data) == _LONG_MESSAGE_SIZE) or
				(report_id == 0x20 and len(data) == _MEDIUM_MESSAGE_SIZE)), \
				"unexpected message size: report_id %02X message %s" % (report_id, _strhex(data))
		if report_id & 0xF0 == 0x00:
			if _log.isEnabledFor(_DEBUG):
				_log.debug("(%s) => r[%02X %s] ignoring unknown report", handle, report_id, _strhex(data[1:]))
			return
		devnumber = ord(data[1:2])

		if _log.isEnabledFor(_DEBUG):
			_log.debug("(%s) => r[%02X %02X %s %s]", handle, report_id, devnumber, _strhex(data[2:4]), _strhex(data[4:]))

		return report_id, devnumber, data[2:]

#
#
#

def _skip_incoming(handle, ihandle, notifications_hook):
	"""Read anything already in the input buffer.

	Used by request() and ping() before their write.
	"""

	while True:
		try:
			# read whatever is already in the buffer, if any
			data = _hid.read(ihandle, _MAX_READ_SIZE, 0)
		except Exception as reason:
			_log.error("read failed, assuming receiver %s no longer available", handle)
			close(handle)
			raise NoReceiver(reason=reason)

		if data:
			assert isinstance(data, bytes), (repr(data), type(data))
			report_id = ord(data[:1])
			if _log.isEnabledFor(_DEBUG):
				assert ((report_id & 0xF0 == 0) or
						(report_id == 0x10 and len(data) == _SHORT_MESSAGE_SIZE) or
						(report_id == 0x11 and len(data) == _LONG_MESSAGE_SIZE) or
						(report_id == 0x20 and len(data) == _MEDIUM_MESSAGE_SIZE)), \
						"unexpected message size: report_id %02X message %s" % (report_id, _strhex(data))
			if notifications_hook and report_id & 0xF0:
				n = make_notification(ord(data[1:2]), data[2:])
				if n:
					notifications_hook(n)
		else:
			# nothing in the input buffer, we're done
			return


def make_notification(devnumber, data):
	"""Guess if this is a notification (and not just a request reply), and
	return a Notification tuple if it is."""
	sub_id = ord(data[:1])
	if sub_id & 0x80 == 0x80:
		# this is either a HID++1.0 register r/w, or an error reply
		return

	address = ord(data[1:2])
	if (
		# standard HID++ 1.0 notification, SubId may be 0x40 - 0x7F
		(sub_id >= 0x40)
		or
		# custom HID++1.0 battery events, where SubId is 0x07/0x0D
		(sub_id in (0x07, 0x0D) and len(data) == 5 and data[4:5] == b'\x00')
		or
		# custom HID++1.0 illumination event, where SubId is 0x17
		(sub_id == 0x17 and len(data) == 5)
		or
		# HID++ 2.0 feature notifications have the SoftwareID 0
		(address & 0x0F == 0x00)
		):
		return _HIDPP_Notification(devnumber, sub_id, address, data[2:])

from collections import namedtuple
_HIDPP_Notification = namedtuple('_HIDPP_Notification', ('devnumber', 'sub_id', 'address', 'data'))
_HIDPP_Notification.__str__ = lambda self: 'Notification(%d,%02X,%02X,%s)' % (self.devnumber, self.sub_id, self.address, _strhex(self.data))
_HIDPP_Notification.__unicode__ = _HIDPP_Notification.__str__
del namedtuple

#
#
#

def request(handle, devnumber, request_id, *params):
	"""Makes a feature call to a device and waits for a matching reply.

	This function will wait for a matching reply indefinitely.

	:param handle: an open UR handle.
	:param devnumber: attached device number.
	:param request_id: a 16-bit integer.
	:param params: parameters for the feature call, 3 to 16 bytes.
	:returns: the reply data, or ``None`` if some error occured.
	"""

	# import inspect as _inspect
	# print ('\n  '.join(str(s) for s in _inspect.stack()))

	assert isinstance(request_id, int)
	if devnumber != 0xFF and request_id < 0x8000:
		# For HID++ 2.0 feature requests, randomize the SoftwareId to make it
		# easier to recognize the reply for this request. also, always set the
		# most significant bit (8) in SoftwareId, to make notifications easier
		# to distinguish from request replies.
		# This only applies to peripheral requests, ofc.
		request_id = (request_id & 0xFFF0) | 0x08 | _random_bits(3)

	timeout = _RECEIVER_REQUEST_TIMEOUT if devnumber == 0xFF else _DEVICE_REQUEST_TIMEOUT
	# be extra patient on long register read
	if request_id & 0xFF00 == 0x8300:
		timeout *= 2

	if params:
		params = b''.join(_pack('B', p) if isinstance(p, int) else p for p in params)
	else:
		params = b''
	# if _log.isEnabledFor(_DEBUG):
	# 	_log.debug("(%s) device %d request_id {%04X} params [%s]", handle, devnumber, request_id, _strhex(params))
	request_data = _pack('!H', request_id) + params

	ihandle = int(handle)
	notifications_hook = getattr(handle, 'notifications_hook', None)
	_skip_incoming(handle, ihandle, notifications_hook)
	write(ihandle, devnumber, request_data)

	# we consider timeout from this point
	request_started = _timestamp()
	delta = 0

	while delta < timeout:
		reply = _read(handle, timeout)

		if reply:
			report_id, reply_devnumber, reply_data = reply
			if reply_devnumber == devnumber:
				if report_id == 0x10 and reply_data[:1] == b'\x8F' and reply_data[1:3] == request_data[:2]:
					error = ord(reply_data[3:4])

					# if error == _hidpp10.ERROR.resource_error: # device unreachable
					# 	_log.warn("(%s) device %d error on request {%04X}: unknown device", handle, devnumber, request_id)
					# 	raise DeviceUnreachable(number=devnumber, request=request_id)

					# if error == _hidpp10.ERROR.unknown_device: # unknown device
					# 	_log.error("(%s) device %d error on request {%04X}: unknown device", handle, devnumber, request_id)
					# 	raise NoSuchDevice(number=devnumber, request=request_id)

					if _log.isEnabledFor(_DEBUG):
						_log.debug("(%s) device 0x%02X error on request {%04X}: %d = %s",
										handle, devnumber, request_id, error, _hidpp10.ERROR[error])
					return

				if reply_data[:1] == b'\xFF' and reply_data[1:3] == request_data[:2]:
					# a HID++ 2.0 feature call returned with an error
					error = ord(reply_data[3:4])
					_log.error("(%s) device %d error on feature request {%04X}: %d = %s",
									handle, devnumber, request_id, error, _hidpp20.ERROR[error])
					raise _hidpp20.FeatureCallError(number=devnumber, request=request_id, error=error, params=params)

				if reply_data[:2] == request_data[:2]:
					if request_id & 0xFE00 == 0x8200:
						# long registry r/w should return a long reply
						assert report_id == 0x11
					elif request_id & 0xFE00 == 0x8000:
						# short registry r/w should return a short reply
						assert report_id == 0x10

					if devnumber == 0xFF:
						if request_id == 0x83B5 or request_id == 0x81F1:
							# these replies have to match the first parameter as well
							if reply_data[2:3] == params[:1]:
								return reply_data[2:]
							else:
								# hm, not mathing my request, and certainly not a notification
								continue
						else:
							return reply_data[2:]
					else:
						return reply_data[2:]
			else:
				# a reply was received, but did not match our request in any way
				# reset the timeout starting point
				request_started = _timestamp()

			if notifications_hook:
				n = make_notification(reply_devnumber, reply_data)
				if n:
					notifications_hook(n)
				# elif _log.isEnabledFor(_DEBUG):
				# 	_log.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))
			# elif _log.isEnabledFor(_DEBUG):
			# 	_log.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))

		delta = _timestamp() - request_started
		# if _log.isEnabledFor(_DEBUG):
		# 	_log.debug("(%s) still waiting for reply, delta %f", handle, delta)

	_log.warn("timeout (%0.2f/%0.2f) on device %d request {%04X} params [%s]",
					delta, timeout, devnumber, request_id, _strhex(params))
	# raise DeviceUnreachable(number=devnumber, request=request_id)


def ping(handle, devnumber):
	"""Check if a device is connected to the receiver.

	:returns: The HID protocol supported by the device, as a floating point number, if the device is active.
	"""
	if _log.isEnabledFor(_DEBUG):
		_log.debug("(%s) pinging device %d", handle, devnumber)

	# import inspect as _inspect
	# print ('\n  '.join(str(s) for s in _inspect.stack()))

	assert devnumber != 0xFF
	assert devnumber > 0x00
	assert devnumber < 0x0F

	# randomize the SoftwareId and mark byte to be able to identify the ping
	# reply, and set most significant (0x8) bit in SoftwareId so that the reply
	# is always distinguishable from notifications
	request_id = 0x0018 | _random_bits(3)
	request_data = _pack('!HBBB', request_id, 0, 0, _random_bits(8))

	ihandle = int(handle)
	notifications_hook = getattr(handle, 'notifications_hook', None)
	_skip_incoming(handle, ihandle, notifications_hook)
	write(ihandle, devnumber, request_data)

	# we consider timeout from this point
	request_started = _timestamp()
	delta = 0

	while delta < _PING_TIMEOUT:
		reply = _read(handle, _PING_TIMEOUT)

		if reply:
			report_id, reply_devnumber, reply_data = reply
			if reply_devnumber == devnumber:
				if reply_data[:2] == request_data[:2] and reply_data[4:5] == request_data[-1:]:
					# HID++ 2.0+ device, currently connected
					return ord(reply_data[2:3]) + ord(reply_data[3:4]) / 10.0

				if report_id == 0x10 and reply_data[:1] == b'\x8F' and reply_data[1:3] == request_data[:2]:
					assert reply_data[-1:] == b'\x00'
					error = ord(reply_data[3:4])

					if error == _hidpp10.ERROR.invalid_SubID__command: # a valid reply from a HID++ 1.0 device
						return 1.0

					if error == _hidpp10.ERROR.resource_error: # device unreachable
						return

					if error == _hidpp10.ERROR.unknown_device: # no paired device with that number
						_log.error("(%s) device %d error on ping request: unknown device", handle, devnumber)
						raise NoSuchDevice(number=devnumber, request=request_id)

			if notifications_hook:
				n = make_notification(reply_devnumber, reply_data)
				if n:
					notifications_hook(n)
				# elif _log.isEnabledFor(_DEBUG):
				# 	_log.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))

		delta = _timestamp() - request_started

	_log.warn("(%s) timeout (%0.2f/%0.2f) on device %d ping", handle, delta, _PING_TIMEOUT, devnumber)
	# raise DeviceUnreachable(number=devnumber, request=request_id)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from time import time as _timestamp

from logitech_receiver import (
				base as _base,
				hidpp10 as _hidpp10,
				status as _status,
				notifications as _notifications,
			)


def run(receivers, args, find_receiver, _ignore):
	assert receivers

	if args.receiver:
		receiver_name = args.receiver.lower()
		receiver = find_receiver(receiver_name)
		if not receiver:
			raise Exception("no receiver found matching '%s'" % receiver_name)
	else:
		receiver = receivers[0]

	assert receiver
	receiver.status = _status.ReceiverStatus(receiver, lambda *args, **kwargs: None)

	# check if it's necessary to set the notification flags
	old_notification_flags = _hidpp10.get_notification_flags(receiver) or 0
	if not (old_notification_flags & _hidpp10.NOTIFICATION_FLAG.wireless):
		_hidpp10.set_notification_flags(receiver, old_notification_flags | _hidpp10.NOTIFICATION_FLAG.wireless)

	# get all current devices
	known_devices = [dev.number for dev in receiver]

	class _HandleWithNotificationHook(int):
		def notifications_hook(self, n):
			assert n
			if n.devnumber == 0xFF:
				_notifications.process(receiver, n)
			elif n.sub_id == 0x41 and n.address == 0x04:
				if n.devnumber not in known_devices:
					receiver.status.new_device = receiver[n.devnumber]

	timeout = 20  # seconds
	receiver.handle = _HandleWithNotificationHook(receiver.handle)

	receiver.set_lock(False, timeout=timeout)
	print ('Pairing: turn your new device on (timing out in', timeout, 'seconds).')

	# the lock-open notification may come slightly later, wait for it a bit
	pairing_start = _timestamp()
	patience = 5  # seconds

	while receiver.status.lock_open or _timestamp() - pairing_start < patience:
		n = _base.read(receiver.handle)
		if n:
			n = _base.make_notification(*n)
			if n:
				receiver.handle.notifications_hook(n)

	if not (old_notification_flags & _hidpp10.NOTIFICATION_FLAG.wireless):
		# only clear the flags if they weren't set before, otherwise a
		# concurrently running Solaar app might stop working properly
		_hidpp10.set_notification_flags(receiver, old_notification_flags)

	if receiver.status.new_device:
		dev = receiver.status.new_device
		print ('Paired device %d: %s (%s) [%s:%s]' % (dev.number, dev.name, dev.codename, dev.wpid, dev.serial))
	else:
		error = receiver.status.get(_status.KEYS.ERROR) or 'no device detected?'
		raise Exception("pairing failed: %s" % error)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, INFO as _INFO
_log = getLogger(__name__)
del getLogger


from solaar.i18n import _
from . import configuration
from logitech_receiver import (
				Receiver,
				listener as _listener,
				status as _status,
				notifications as _notifications
			)

#
#
#

from collections import namedtuple
_GHOST_DEVICE = namedtuple('_GHOST_DEVICE', ('receiver', 'number', 'name', 'kind', 'status', 'online'))
_GHOST_DEVICE.__bool__ = lambda self: False
_GHOST_DEVICE.__nonzero__ = _GHOST_DEVICE.__bool__
del namedtuple

def _ghost(device):
	return _GHOST_DEVICE(
					receiver=device.receiver,
					number=device.number,
					name=device.name,
					kind=device.kind,
					status=None,
					online=False)

#
#
#

# how often to poll devices that haven't updated their statuses on their own
# (through notifications)
# _POLL_TICK = 5 * 60  # seconds


class ReceiverListener(_listener.EventsListener):
	"""Keeps the status of a Receiver.
	"""
	def __init__(self, receiver, status_changed_callback):
		super(ReceiverListener, self).__init__(receiver, self._notifications_handler)
		# no reason to enable polling yet
		# self.tick_period = _POLL_TICK
		# self._last_tick = 0

		assert status_changed_callback
		self.status_changed_callback = status_changed_callback
		_status.attach_to(receiver, self._status_changed)

	def has_started(self):
		if _log.isEnabledFor(_INFO):
			_log.info("%s: notifications listener has started (%s)", self.receiver, self.receiver.handle)
		notification_flags = self.receiver.enable_notifications()
		self.receiver.status[_status.KEYS.NOTIFICATION_FLAGS] = notification_flags
		self.receiver.notify_devices()
		self._status_changed(self.receiver)  #, _status.ALERT.NOTIFICATION)

	def has_stopped(self):
		r, self.receiver = self.receiver, None
		assert r is not None
		if _log.isEnabledFor(_INFO):
			_log.info("%s: notifications listener has stopped", r)

		# because udev is not notifying us about device removal,
		# make sure to clean up in _all_listeners
		_all_listeners.pop(r.path, None)

		r.status = _("The receiver was unplugged.")
		if r:
			try:
				r.close()
			except:
				_log.exception("closing receiver %s" % r.path)
		self.status_changed_callback(r)  #, _status.ALERT.NOTIFICATION)

	# def tick(self, timestamp):
	# 	if not self.tick_period:
	# 		raise Exception("tick() should not be called without a tick_period: %s", self)
	#
	# 	# not necessary anymore, we're now using udev monitor to watch for receiver status
	# 	# if self._last_tick > 0 and timestamp - self._last_tick > _POLL_TICK * 2:
	# 	# 	# if we missed a couple of polls, most likely the computer went into
	# 	# 	# sleep, and we have to reinitialize the receiver again
	# 	# 	_log.warn("%s: possible sleep detected, closing this listener", self.receiver)
	# 	# 	self.stop()
	# 	# 	return
	#
	# 	self._last_tick = timestamp
	#
	# 	try:
	# 		# read these in case they haven't been read already
	# 		# self.receiver.serial, self.receiver.firmware
	# 		if self.receiver.status.lock_open:
	# 			# don't mess with stuff while pairing
	# 			return
	#
	# 		self.receiver.status.poll(timestamp)
	#
	# 		# Iterating directly through the reciver would unnecessarily probe
	# 		# all possible devices, even unpaired ones.
	# 		# Checking for each device number in turn makes sure only already
	# 		# known devices are polled.
	# 		# This is okay because we should have already known about them all
	# 		# long before the first poll() happents, through notifications.
	# 		for number in range(1, 6):
	# 			if number in self.receiver:
	# 				dev = self.receiver[number]
	# 				if dev and dev.status is not None:
	# 					dev.status.poll(timestamp)
	# 	except Exception as e:
	# 		_log.exception("polling", e)

	def _status_changed(self, device, alert=_status.ALERT.NONE, reason=None):
		assert device is not None
		if _log.isEnabledFor(_INFO):
			if device.kind is None:
				_log.info("status_changed %s: %s, %s (%X) %s", device,
							'present' if bool(device) else 'removed',
							device.status, alert, reason or '')
			else:
				_log.info("status_changed %s: %s %s, %s (%X) %s", device,
							'paired' if bool(device) else 'unpaired',
							'online' if device.online else 'offline',
							device.status, alert, reason or '')

		if device.kind is None:
			assert device == self.receiver
			# the status of the receiver changed
			self.status_changed_callback(device, alert, reason)
			return

		assert device.receiver == self.receiver
		if not device:
			# Device was unpaired, and isn't valid anymore.
			# We replace it with a ghost so that the UI has something to work
			# with while cleaning up.
			_log.warn("device %s was unpaired, ghosting", device)
			device = _ghost(device)

		self.status_changed_callback(device, alert, reason)

		if not device:
			# the device was just unpaired, need to update the
			# status of the receiver as well
			self.status_changed_callback(self.receiver)

	def _notifications_handler(self, n):
		assert self.receiver
		# if _log.isEnabledFor(_DEBUG):
		# 	_log.debug("%s: handling %s", self.receiver, n)
		if n.devnumber == 0xFF:
			# a receiver notification
			_notifications.process(self.receiver, n)
			return

		# a device notification
		assert n.devnumber > 0 and n.devnumber <= self.receiver.max_devices
		already_known = n.devnumber in self.receiver
		if not already_known and n.sub_id == 0x41:
			dev = self.receiver.register_new_device(n.devnumber, n)
		else:
			dev = self.receiver[n.devnumber]

		if not dev:
			_log.warn("%s: received %s for invalid device %d: %r", self.receiver, n, n.devnumber, dev)
			return

		if not already_known:
			if _log.isEnabledFor(_INFO):
				_log.info("%s triggered new device %s (%s)", n, dev, dev.kind)
			# If there are saved configs, bring the device's settings up-to-date.
			# They will be applied when the device is marked as online.
			configuration.attach_to(dev)
			_status.attach_to(dev, self._status_changed)
			# the receiver changed status as well
			self._status_changed(self.receiver)

		assert dev
		assert dev.status is not None
		_notifications.process(dev, n)
		if self.receiver.status.lock_open and not already_known:
			# this should be the first notification after a device was paired
			assert n.sub_id == 0x41 and n.address == 0x04
			if _log.isEnabledFor(_INFO):
				_log.info("%s: pairing detected new device", self.receiver)
			self.receiver.status.new_device = dev
		elif dev:
			if dev.online is None:
				dev.ping()

	def __str__(self):
		return '<ReceiverListener(%s,%s)>' % (self.receiver.path, self.receiver.handle)
	__unicode__ = __str__

#
#
#

# all known receiver listeners
# listeners that stop on their own may remain here
_all_listeners = {}


def _start(device_info):
	assert _status_callback
	receiver = Receiver.open(device_info)
	if receiver:
		rl = ReceiverListener(receiver, _status_callback)
		rl.start()
		_all_listeners[device_info.path] = rl
		return rl

	_log.warn("failed to open %s", device_info)


def start_all():
	# just in case this it called twice in a row...
	stop_all()

	if _log.isEnabledFor(_INFO):
		_log.info("starting receiver listening threads")
	for device_info in _base.receivers():
		_process_receiver_event('add', device_info)


def stop_all():
	listeners = list(_all_listeners.values())
	_all_listeners.clear()

	if listeners:
		if _log.isEnabledFor(_INFO):
			_log.info("stopping receiver listening threads %s", listeners)

		for l in listeners:
			l.stop()

	configuration.save()

	if listeners:
		for l in listeners:
			l.join()


def ping_all():
	for l in _all_listeners.values():
		count = l.receiver.count()
		if count:
			for dev in l.receiver:
				dev.ping()
				l._status_changed(dev)
				count -= 1
				if not count:
					break


from logitech_receiver import base as _base
_status_callback = None
_error_callback = None

def setup_scanner(status_changed_callback, error_callback):
	global _status_callback, _error_callback
	assert _status_callback is None, 'scanner was already set-up'

	_status_callback = status_changed_callback
	_error_callback = error_callback

	_base.notify_on_receivers_glib(_process_receiver_event)


# receiver add/remove events will start/stop listener threads
def _process_receiver_event(action, device_info):
	assert action is not None
	assert device_info is not None
	assert _error_callback

	if _log.isEnabledFor(_INFO):
		_log.info("receiver event %s %s", action, device_info)

	# whatever the action, stop any previous receivers at this path
	l = _all_listeners.pop(device_info.path, None)
	if l is not None:
		assert isinstance(l, ReceiverListener)
		l.stop()

	if action == 'add':
		# a new receiver device was detected
		try:
			_start(device_info)
		except OSError:
			# permission error, ignore this path for now
			_error_callback('permissions', device_info.path)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Translation support for the Logitech receivers library

from __future__ import absolute_import, division, print_function, unicode_literals

import gettext as _gettext


try:
	unicode
	_ = lambda x: _gettext.gettext(x).decode('UTF-8')
except:
	_ = _gettext.gettext


# A few common strings, not always accessible as such in the code.

_DUMMY = (
		# approximative battery levels
		_("empty"), _("critical"), _("low"), _("good"), _("full"),

		# battery charging statuses
		_("discharging"), _("recharging"), _("almost full"), _("full"),
		_("slow recharge"), _("invalid battery"), _("thermal error"),

		# pairing errors
		_("device timeout"), _("device not supported"), _("too many devices"), _("sequence timeout"),

		# firmware kinds
		_("Firmware"), _("Bootloader"), _("Hardware"), _("Other"),

	)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from logitech_receiver import (
				hidpp10 as _hidpp10,
				hidpp20 as _hidpp20,
				special_keys as _special_keys,
			)


def _print_receiver(receiver):
	paired_count = receiver.count()

	print ('Unifying Receiver')
	print ('  Device path  :', receiver.path)
	print ('  USB id       : 046d:%s' % receiver.product_id)
	print ('  Serial       :', receiver.serial)
	for f in receiver.firmware:
		print ('    %-11s: %s' % (f.kind, f.version))

	print ('  Has', paired_count, 'paired device(s) out of a maximum of %d.' % receiver.max_devices)

	notification_flags = _hidpp10.get_notification_flags(receiver)
	if notification_flags is not None:
		if notification_flags:
			notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
			print ('  Notifications: %s (0x%06X)' % (', '.join(notification_names), notification_flags))
		else:
			print ('  Notifications: (none)')

	activity = receiver.read_register(_hidpp10.REGISTERS.devices_activity)
	if activity:
		activity = [(d, ord(activity[d - 1:d])) for d in range(1, receiver.max_devices)]
		activity_text = ', '.join(('%d=%d' % (d, a)) for d, a in activity if a > 0)
		print ('  Device activity counters:', activity_text or '(empty)')


def _print_device(dev):
	assert dev
	# check if the device is online
	dev.ping()

	print ('  %d: %s' % (dev.number, dev.name))
	print ('     Codename     :', dev.codename)
	print ('     Kind         :', dev.kind)
	print ('     Wireless PID :', dev.wpid)
	if dev.protocol:
		print ('     Protocol     : HID++ %1.1f' % dev.protocol)
	else:
		print ('     Protocol     : unknown (device is offline)')
	if dev.polling_rate:
		print ('     Polling rate :', dev.polling_rate, 'ms (%dHz)' % (1000 // dev.polling_rate))
	print ('     Serial number:', dev.serial)
	for fw in dev.firmware:
		print ('       %11s:' % fw.kind, (fw.name + ' ' + fw.version).strip())

	if dev.power_switch_location:
		print ('     The power switch is located on the %s.' % dev.power_switch_location)

	if dev.online:
		notification_flags = _hidpp10.get_notification_flags(dev)
		if notification_flags is not None:
			if notification_flags:
				notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
				print ('     Notifications: %s (0x%06X).' % (', '.join(notification_names), notification_flags))
			else:
				print ('     Notifications: (none).')

	if dev.online and dev.features:
		print ('     Supports %d HID++ 2.0 features:' % len(dev.features))
		for index, feature in enumerate(dev.features):
			feature = dev.features[index]
			flags = dev.request(0x0000, feature.bytes(2))
			flags = 0 if flags is None else ord(flags[1:2])
			flags = _hidpp20.FEATURE_FLAG.flag_names(flags)
			print ('        %2d: %-22s {%04X}   %s' % (index, feature, feature, ', '.join(flags)))

	if dev.online and dev.keys:
		print ('     Has %d reprogrammable keys:' % len(dev.keys))
		for k in dev.keys:
			flags = _special_keys.KEY_FLAG.flag_names(k.flags)
			print ('        %2d: %-26s => %-27s   %s' % (k.index, k.key, k.task, ', '.join(flags)))

	if dev.online:
		battery = _hidpp20.get_battery(dev)
		if battery is None:
			battery = _hidpp10.get_battery(dev)
		if battery is not None:
			from logitech_receiver.common import NamedInt as _NamedInt
			level, status = battery
			if level is not None:
				if isinstance(level, _NamedInt):
					text = str(level)
				else:
					text = '%d%%' % level
			else:
				text = 'N/A'
			print ('     Battery: %s, %s.' % (text, status))
		else:
			print ('     Battery status unavailable.')
	else:
		print ('     Battery: unknown (device is offline).')


def run(receivers, args, find_receiver, find_device):
	assert receivers
	assert args.device

	device_name = args.device.lower()

	if device_name == 'all':
		for r in receivers:
			_print_receiver(r)
			count = r.count()
			if count:
				for dev in r:
					print ('')
					_print_device(dev)
					count -= 1
					if not count:
						break
			print ('')
		return

	dev = find_receiver(receivers, device_name)
	if dev:
		_print_receiver(dev)
		return

	dev = find_device(receivers, device_name)
	assert dev
	_print_device(dev)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, INFO as _INFO
_log = getLogger(__name__)
del getLogger

#
# As suggested here: http://stackoverflow.com/a/13548984
#

_suspend_callback = None
def _suspend():
	if _suspend_callback:
		if _log.isEnabledFor(_INFO):
			_log.info("received suspend event from UPower")
		_suspend_callback()


_resume_callback = None
def _resume():
	if _resume_callback:
		if _log.isEnabledFor(_INFO):
			_log.info("received resume event from UPower")
		_resume_callback()


def watch(on_resume_callback=None, on_suspend_callback=None):
	"""Register callback for suspend/resume events.
	They are called only if the system DBus is running, and the UPower daemon is available."""
	global _resume_callback, _suspend_callback
	_suspend_callback = on_suspend_callback
	_resume_callback = on_resume_callback


try:
	import dbus

	_UPOWER_BUS = 'org.freedesktop.UPower'
	_UPOWER_INTERFACE = 'org.freedesktop.UPower'

	# integration into the main GLib loop
	from dbus.mainloop.glib import DBusGMainLoop
	DBusGMainLoop(set_as_default=True)

	bus = dbus.SystemBus()
	assert bus

	bus.add_signal_receiver(_suspend, signal_name='Sleeping',
					dbus_interface=_UPOWER_INTERFACE, bus_name=_UPOWER_BUS)

	bus.add_signal_receiver(_resume, signal_name='Resuming',
					dbus_interface=_UPOWER_INTERFACE, bus_name=_UPOWER_BUS)

	if _log.isEnabledFor(_INFO):
		_log.info("connected to system dbus, watching for suspend/resume events")

except:
	# Either:
	# - the dbus library is not available
	# - the system dbus is not running
	_log.warn("failed to register suspend/resume callbacks")
	pass



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from solaar import configuration as _configuration
from logitech_receiver import settings as _settings


def _print_setting(s, verbose=True):
	print ('#', s.label)
	if verbose:
		print ('#', s.description.replace('\n', ' '))
		if s.kind == _settings.KIND.toggle:
			print ('#   possible values: on/true/t/yes/y/1 or off/false/f/no/n/0')
		elif s.choices:
			print ('#   possible values: one of [', ', '.join(str(v) for v in s.choices), '], or higher/lower/highest/max/lowest/min')
		else:
			# wtf?
			pass
	value = s.read(cached=False)
	if value is None:
		print (s.name, '= ? (failed to read from device)')
	else:
		print (s.name, '= %r' % value)


def run(receivers, args, find_receiver, find_device):
	assert receivers
	assert args.device

	device_name = args.device.lower()
	dev = find_device(receivers, device_name)

	if not dev.ping():
		raise Exception('%s is offline' % dev.name)

	if not dev.settings:
		raise Exception('no settings for %s' % dev.name)

	_configuration.attach_to(dev)

	if not args.setting:
		print (dev.name, '(%s) [%s:%s]' % (dev.codename, dev.wpid, dev.serial))
		for s in dev.settings:
			print ('')
			_print_setting(s)
		return

	setting_name = args.setting.lower()
	setting = None
	for s in dev.settings:
		if setting_name == s.name.lower():
			setting = s
			break
	if setting is None:
		raise Exception("no setting '%s' for %s" % (args.setting, dev.name))

	if args.value is None:
		_print_setting(setting)
		return

	if setting.kind == _settings.KIND.toggle:
		value = args.value
		try:
			value = bool(int(value))
		except:
			if value.lower() in ('true', 'yes', 'on', 't', 'y'):
				value = True
			elif value.lower() in ('false', 'no', 'off', 'f', 'n'):
				value = False
			else:
				raise Exception("don't know how to interpret '%s' as boolean" % value)

	elif setting.choices:
		value = args.value.lower()

		if value in ('higher', 'lower'):
			old_value = setting.read()
			if old_value is None:
				raise Exception("could not read current value of '%s'" % setting.name)

			if value == 'lower':
				lower_values = setting.choices[:old_value]
				value = lower_values[-1] if lower_values else setting.choices[:][0]
			elif value == 'higher':
				higher_values = setting.choices[old_value + 1:]
				value = higher_values[0] if higher_values else setting.choices[:][-1]
		elif value in ('highest', 'max'):
			value = setting.choices[:][-1]
		elif value in ('lowest', 'min'):
			value = setting.choices[:][0]
		elif value not in setting.choices:
			raise Exception("possible values for '%s' are: [%s]" % (setting.name, ', '.join(str(v) for v in setting.choices)))
			value = setting.choices[value]

	else:
		raise NotImplemented

	result = setting.write(value)
	if result is None:
		raise Exception("failed to set '%s' = '%s' [%r]" % (setting.name, value, value))
	_print_setting(setting, False)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Handles incoming events from the receiver/devices, updating the related
# status object as appropiate.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, DEBUG as _DEBUG, INFO as _INFO
_log = getLogger(__name__)
del getLogger


from .i18n import _
from .common import strhex as _strhex, unpack as _unpack
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .status import KEYS as _K, ALERT as _ALERT

_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
#
#

def process(device, notification):
	assert device
	assert notification

	assert hasattr(device, 'status')
	status = device.status
	assert status is not None

	if device.kind is None:
		return _process_receiver_notification(device, status, notification)

	return _process_device_notification(device, status, notification)

#
#
#

def _process_receiver_notification(receiver, status, n):
	# supposedly only 0x4x notifications arrive for the receiver
	assert n.sub_id & 0x40 == 0x40

	# pairing lock notification
	if n.sub_id == 0x4A:
		status.lock_open = bool(n.address & 0x01)
		reason = _("pairing lock is ") + (_("open") if status.lock_open else _("closed"))
		if _log.isEnabledFor(_INFO):
			_log.info("%s: %s", receiver, reason)

		status[_K.ERROR] = None
		if status.lock_open:
			status.new_device = None

		pair_error = ord(n.data[:1])
		if pair_error:
			status[_K.ERROR] = error_string = _hidpp10.PAIRING_ERRORS[pair_error]
			status.new_device = None
			_log.warn("pairing error %d: %s", pair_error, error_string)

		status.changed(reason=reason)
		return True

	_log.warn("%s: unhandled notification %s", receiver, n)

#
#
#

def _process_device_notification(device, status, n):
	# incoming packets with SubId >= 0x80 are supposedly replies from
	# HID++ 1.0 requests, should never get here
	assert n.sub_id & 0x80 == 0

	# 0x40 to 0x7F appear to be HID++ 1.0 notifications
	if n.sub_id >= 0x40:
		return _process_hidpp10_notification(device, status, n)

	# At this point, we need to know the device's protocol, otherwise it's
	# possible to not know how to handle it.
	assert device.protocol is not None

	# some custom battery events for HID++ 1.0 devices
	if device.protocol < 2.0:
		return _process_hidpp10_custom_notification(device, status, n)

	# assuming 0x00 to 0x3F are feature (HID++ 2.0) notifications
	assert device.features
	try:
		feature = device.features[n.sub_id]
	except IndexError:
		_log.warn("%s: notification from invalid feature index %02X: %s", device, n.sub_id, n)
		return False

	return _process_feature_notification(device, status, n, feature)


def _process_hidpp10_custom_notification(device, status, n):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("%s (%s) custom notification %s", device, device.protocol, n)

	if n.sub_id in (_R.battery_status, _R.battery_charge):
		# message layout: 10 ix <register> <xx> <yy> <zz> <00>
		assert n.data[-1:] == b'\x00'
		data = chr(n.address).encode() + n.data
		charge, status_text = _hidpp10.parse_battery_status(n.sub_id, data)
		status.set_battery_info(charge, status_text)
		return True

	if n.sub_id == _R.keyboard_illumination:
		# message layout: 10 ix 17("address")  <??> <?> <??> <light level 1=off..5=max>
		# TODO anything we can do with this?
		if _log.isEnabledFor(_INFO):
			_log.info("illumination event: %s", n)
		return True

	_log.warn("%s: unrecognized %s", device, n)


def _process_hidpp10_notification(device, status, n):
	# unpair notification
	if n.sub_id == 0x40:
		if n.address == 0x02:
			# device un-paired
			status.clear()
			device.wpid = None
			device.status = None
			if device.number in device.receiver:
				del device.receiver[device.number]
			status.changed(active=False, alert=_ALERT.ALL, reason=_("unpaired"))
		else:
			_log.warn("%s: disconnection with unknown type %02X: %s", device, n.address, n)
		return True

	# wireless link notification
	if n.sub_id == 0x41:
		protocol_name = ('unifying (eQuad DJ)' if n.address == 0x04
					else 'eQuad' if n.address == 0x03
					else None)
		if protocol_name:
			if _log.isEnabledFor(_DEBUG):
				wpid = _strhex(n.data[2:3] + n.data[1:2])
				assert wpid == device.wpid, "%s wpid mismatch, got %s" % (device, wpid)

			flags = ord(n.data[:1]) & 0xF0
			link_encrypyed = bool(flags & 0x20)
			link_established = not (flags & 0x40)
			if _log.isEnabledFor(_DEBUG):
				sw_present = bool(flags & 0x10)
				has_payload = bool(flags & 0x80)
				_log.debug("%s: %s connection notification: software=%s, encrypted=%s, link=%s, payload=%s",
							device, protocol_name, sw_present, link_encrypyed, link_established, has_payload)
			status[_K.LINK_ENCRYPTED] = link_encrypyed
			status.changed(active=link_established)
		else:
			_log.warn("%s: connection notification with unknown protocol %02X: %s", device.number, n.address, n)

		return True

	if n.sub_id == 0x49:
		# raw input event? just ignore it
		# if n.address == 0x01, no idea what it is, but they keep on coming
		# if n.address == 0x03, appears to be an actual input event,
		#     because they only come when input happents
		return True

	# power notification
	if n.sub_id == 0x4B:
		if n.address == 0x01:
			if _log.isEnabledFor(_DEBUG):
				_log.debug("%s: device powered on", device)
			reason = status.to_string() or _("powered on")
			status.changed(active=True, alert=_ALERT.NOTIFICATION, reason=reason)
		else:
			_log.warn("%s: unknown %s", device, n)
		return True

	_log.warn("%s: unrecognized %s", device, n)


def _process_feature_notification(device, status, n, feature):
	if feature == _F.BATTERY_STATUS:
		if n.address == 0x00:
			discharge = ord(n.data[:1])
			battery_status = ord(n.data[1:2])
			status.set_battery_info(discharge, _hidpp20.BATTERY_STATUS[battery_status])
		else:
			_log.warn("%s: unknown BATTERY %s", device, n)
		return True

	# TODO: what are REPROG_CONTROLS_V{2,3}?
	if feature == _F.REPROG_CONTROLS:
		if n.address == 0x00:
			if _log.isEnabledFor(_INFO):
				_log.info("%s: reprogrammable key: %s", device, n)
		else:
			_log.warn("%s: unknown REPROGRAMMABLE KEYS %s", device, n)
		return True

	if feature == _F.WIRELESS_DEVICE_STATUS:
		if n.address == 0x00:
			if _log.isEnabledFor(_DEBUG):
				_log.debug("wireless status: %s", n)
			if n.data[0:3] == b'\x01\x01\x01':
				status.changed(active=True, alert=_ALERT.NOTIFICATION, reason='powered on')
			else:
				_log.warn("%s: unknown WIRELESS %s", device, n)
		else:
			_log.warn("%s: unknown WIRELESS %s", device, n)
		return True

	if feature == _F.SOLAR_DASHBOARD:
		if n.data[5:9] == b'GOOD':
			charge, lux, adc = _unpack('!BHH', n.data[:5])
			# guesstimate the battery voltage, emphasis on 'guess'
			# status_text = '%1.2fV' % (adc * 2.67793237653 / 0x0672)
			status_text = _hidpp20.BATTERY_STATUS.discharging

			if n.address == 0x00:
				status[_K.LIGHT_LEVEL] = None
				status.set_battery_info(charge, status_text)
			elif n.address == 0x10:
				status[_K.LIGHT_LEVEL] = lux
				if lux > 200:
					status_text = _hidpp20.BATTERY_STATUS.recharging
				status.set_battery_info(charge, status_text)
			elif n.address == 0x20:
				if _log.isEnabledFor(_DEBUG):
					_log.debug("%s: Light Check button pressed", device)
				status.changed(alert=_ALERT.SHOW_WINDOW)
				# first cancel any reporting
				# device.feature_request(_F.SOLAR_DASHBOARD)
				# trigger a new report chain
				reports_count = 15
				reports_period = 2  # seconds
				device.feature_request(_F.SOLAR_DASHBOARD, 0x00, reports_count, reports_period)
			else:
				_log.warn("%s: unknown SOLAR CHAGE %s", device, n)
		else:
			_log.warn("%s: SOLAR CHARGE not GOOD? %s", device, n)
		return True

	if feature == _F.TOUCHMOUSE_RAW_POINTS:
		if n.address == 0x00:
			if _log.isEnabledFor(_INFO):
				_log.info("%s: TOUCH MOUSE points %s", device, n)
		elif n.address == 0x10:
			touch = ord(n.data[:1])
			button_down = bool(touch & 0x02)
			mouse_lifted = bool(touch & 0x01)
			if _log.isEnabledFor(_INFO):
				_log.info("%s: TOUCH MOUSE status: button_down=%s mouse_lifted=%s", device, button_down, mouse_lifted)
		else:
			_log.warn("%s: unknown TOUCH MOUSE %s", device, n)
		return True

	_log.warn("%s: unrecognized %s for feature %s (index %02X)", device, n, feature, n.sub_id)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# USB ids of Logitech wireless receivers.
# Only receivers supporting the HID++ protocol can go in here.

from __future__ import absolute_import, division, print_function, unicode_literals


_UNIFYING_DRIVER = 'logitech-djreceiver'
_GENERIC_DRIVER = ('hid-generic', 'generic-usb')


# each tuple contains (vendor_id, product_id, usb interface number, hid driver)
_unifying_receiver = lambda product_id: (0x046d, product_id, 2, _UNIFYING_DRIVER)
_nano_receiver = lambda product_id: (0x046d, product_id, 1, _GENERIC_DRIVER)


# standard Unifying receivers (marked with the orange Unifying logo)
UNIFYING_RECEIVER_C52B    = _unifying_receiver(0xc52b)
UNIFYING_RECEIVER_C532    = _unifying_receiver(0xc532)

# Nano receviers that support the Unifying protocol
NANO_RECEIVER_ADVANCED    = _nano_receiver(0xc52f)

# Nano receivers that don't support the Unifying protocol
NANO_RECEIVER_C517        = _nano_receiver(0xc517)
NANO_RECEIVER_C518        = _nano_receiver(0xc518)
NANO_RECEIVER_C51A        = _nano_receiver(0xc51a)
NANO_RECEIVER_C51B        = _nano_receiver(0xc51b)
NANO_RECEIVER_C521        = _nano_receiver(0xc521)
NANO_RECEIVER_C525        = _nano_receiver(0xc525)
NANO_RECEIVER_C526        = _nano_receiver(0xc526)
NANO_RECEIVER_C52e        = _nano_receiver(0xc52e)


del _unifying_receiver, _nano_receiver


ALL = (
		UNIFYING_RECEIVER_C52B,
		UNIFYING_RECEIVER_C532,
		NANO_RECEIVER_ADVANCED,
		NANO_RECEIVER_C517,
		NANO_RECEIVER_C518,
		NANO_RECEIVER_C51A,
		NANO_RECEIVER_C51B,
		NANO_RECEIVER_C521,
		NANO_RECEIVER_C525,
		NANO_RECEIVER_C526,
		NANO_RECEIVER_C52e,
	)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger  # , DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger


from .common import (strhex as _strhex,
					bytes2int as _bytes2int,
					int2bytes as _int2bytes,
					NamedInts as _NamedInts,
					FirmwareInfo as _FirmwareInfo)
from .hidpp20 import FIRMWARE_KIND, BATTERY_STATUS

#
# Constants - most of them as defined by the official Logitech HID++ 1.0
# documentation, some of them guessed.
#

DEVICE_KIND = _NamedInts(
				keyboard=0x01,
				mouse=0x02,
				numpad=0x03,
				presenter=0x04,
				trackball=0x08,
				touchpad=0x09)

POWER_SWITCH_LOCATION = _NamedInts(
				base=0x01,
				top_case=0x02,
				edge_of_top_right_corner=0x03,
				top_left_corner=0x05,
				bottom_left_corner=0x06,
				top_right_corner=0x07,
				bottom_right_corner=0x08,
				top_edge=0x09,
				right_edge=0x0A,
				left_edge=0x0B,
				bottom_edge=0x0C)

# Some flags are used both by devices and receivers. The Logitech documentation
# mentions that the first and last (third) byte are used for devices while the
# second is used for the receiver. In practise, the second byte is also used for
# some device-specific notifications (keyboard illumination level). Do not
# simply set all notification bits if the software does not support it. For
# example, enabling keyboard_sleep_raw makes the Sleep key a no-operation unless
# the software is updated to handle that event.
# Observations:
# - wireless and software present were seen on receivers, reserved_r1b4 as well
# - the rest work only on devices as far as we can tell right now
# In the future would be useful to have separate enums for receiver and device notification flags,
# but right now we don't know enough.
NOTIFICATION_FLAG = _NamedInts(
				battery_status=         0x100000,  # send battery charge notifications (0x07 or 0x0D)
				keyboard_sleep_raw=     0x020000,  # system control keys such as Sleep
				keyboard_multimedia_raw=0x010000,  # consumer controls such as Mute and Calculator
				# reserved_r1b4=        0x001000,  # unknown, seen on a unifying receiver
				software_present=       0x000800,  # .. no idea
				keyboard_illumination=  0x000200,  # illumination brightness level changes (by pressing keys)
				wireless=               0x000100,  # notify when the device wireless goes on/off-line
				)

ERROR = _NamedInts(
				invalid_SubID__command=0x01,
				invalid_address=0x02,
				invalid_value=0x03,
				connection_request_failed=0x04,
				too_many_devices=0x05,
				already_exists=0x06,
				busy=0x07,
				unknown_device=0x08,
				resource_error=0x09,
				request_unavailable=0x0A,
				unsupported_parameter_value=0x0B,
				wrong_pin_code=0x0C)

PAIRING_ERRORS = _NamedInts(
				device_timeout=0x01,
				device_not_supported=0x02,
				too_many_devices=0x03,
				sequence_timeout=0x06)

BATTERY_APPOX = _NamedInts(
				empty = 0,
				critical = 5,
				low = 20,
				good = 50,
				full = 90)

"""Known registers.
Devices usually have a (small) sub-set of these. Some registers are only
applicable to certain device kinds (e.g. smooth_scroll only applies to mice."""
REGISTERS = _NamedInts(
				# only apply to receivers
				receiver_connection=0x02,
				receiver_pairing=0xB2,
				devices_activity=0x2B3,
				receiver_info=0x2B5,

				# only apply to devices
				mouse_button_flags=0x01,
				keyboard_hand_detection=0x01,
				battery_status=0x07,
				keyboard_fn_swap=0x09,
				battery_charge=0x0D,
				keyboard_illumination=0x17,
				three_leds=0x51,
				mouse_dpi=0x63,

				# apply to both
				notifications=0x00,
				firmware=0xF1,
			)

#
# functions
#

def read_register(device, register_number, *params):
	assert device, 'tried to read register %02X from invalid device %s' % (register_number, device)
	# support long registers by adding a 2 in front of the register number
	request_id = 0x8100 | (int(register_number) & 0x2FF)
	return device.request(request_id, *params)


def write_register(device, register_number, *value):
	assert device, 'tried to write register %02X to invalid device %s' % (register_number, device)
	# support long registers by adding a 2 in front of the register number
	request_id = 0x8000 | (int(register_number) & 0x2FF)
	return device.request(request_id, *value)


def get_battery(device):
	assert device
	assert device.kind is not None
	if not device.online:
		return

	"""Reads a device's battery level, if provided by the HID++ 1.0 protocol."""
	if device.protocol and device.protocol >= 2.0:
		# let's just assume HID++ 2.0 devices do not provide the battery info in a register
		return

	for r in (REGISTERS.battery_status, REGISTERS.battery_charge):
		if r in device.registers:
			reply = read_register(device, r)
			if reply:
				return parse_battery_status(r, reply)
			return

	# the descriptor does not tell us which register this device has, try them both
	reply = read_register(device, REGISTERS.battery_charge)
	if reply:
		# remember this for the next time
		device.registers.append(REGISTERS.battery_charge)
		return parse_battery_status(REGISTERS.battery_charge, reply)

	reply = read_register(device, REGISTERS.battery_status)
	if reply:
		# remember this for the next time
		device.registers.append(REGISTERS.battery_status)
		return parse_battery_status(REGISTERS.battery_status, reply)


def parse_battery_status(register, reply):
	if register == REGISTERS.battery_charge:
		charge = ord(reply[:1])
		status_byte = ord(reply[2:3]) & 0xF0
		status_text = (BATTERY_STATUS.discharging if status_byte == 0x30
				else BATTERY_STATUS.recharging if status_byte == 0x50
				else BATTERY_STATUS.full if status_byte == 0x90
				else None)
		return charge, status_text

	if register == REGISTERS.battery_status:
		status_byte = ord(reply[:1])
		charge = (BATTERY_APPOX.full if status_byte == 7 # full
			else BATTERY_APPOX.good if status_byte == 5 # good
			else BATTERY_APPOX.low if status_byte == 3 # low
			else BATTERY_APPOX.critical if status_byte == 1 # critical
			# pure 'charging' notifications may come without a status
			else BATTERY_APPOX.empty)

		charging_byte = ord(reply[1:2])
		if charging_byte == 0x00:
			status_text = BATTERY_STATUS.discharging
		elif charging_byte & 0x21 == 0x21:
			status_text = BATTERY_STATUS.recharging
		elif charging_byte & 0x22 == 0x22:
			status_text = BATTERY_STATUS.full
		else:
			_log.warn("could not parse 0x07 battery status: %02X (level %02X)", charging_byte, status_byte)
			status_text = None

		if charging_byte & 0x03 and status_byte == 0:
			# some 'charging' notifications may come with no battery level information
			charge = None

		return charge, status_text


def get_firmware(device):
	assert device

	firmware = [None, None, None]

	reply = read_register(device, REGISTERS.firmware, 0x01)
	if not reply:
		# won't be able to read any of it now...
		return

	fw_version = _strhex(reply[1:3])
	fw_version = '%s.%s' % (fw_version[0:2], fw_version[2:4])
	reply = read_register(device, REGISTERS.firmware, 0x02)
	if reply:
		fw_version += '.B' + _strhex(reply[1:3])
	fw = _FirmwareInfo(FIRMWARE_KIND.Firmware, '', fw_version, None)
	firmware[0] = fw

	reply = read_register(device, REGISTERS.firmware, 0x04)
	if reply:
		bl_version = _strhex(reply[1:3])
		bl_version = '%s.%s' % (bl_version[0:2], bl_version[2:4])
		bl = _FirmwareInfo(FIRMWARE_KIND.Bootloader, '', bl_version, None)
		firmware[1] = bl

	reply = read_register(device, REGISTERS.firmware, 0x03)
	if reply:
		o_version = _strhex(reply[1:3])
		o_version = '%s.%s' % (o_version[0:2], o_version[2:4])
		o = _FirmwareInfo(FIRMWARE_KIND.Other, '', o_version, None)
		firmware[2] = o

	if any(firmware):
		return tuple(f for f in firmware if f)


def set_3leds(device, battery_level=None, charging=None, warning=None):
	assert device
	assert device.kind is not None
	if not device.online:
		return

	if REGISTERS.three_leds not in device.registers:
		return

	if battery_level is not None:
		if battery_level < BATTERY_APPOX.critical:
			# 1 orange, and force blink
			v1, v2 = 0x22, 0x00
			warning = True
		elif battery_level < BATTERY_APPOX.low:
			# 1 orange
			v1, v2 = 0x22, 0x00
		elif battery_level < BATTERY_APPOX.good:
			# 1 green
			v1, v2 = 0x20, 0x00
		elif battery_level < BATTERY_APPOX.full:
			# 2 greens
			v1, v2 = 0x20, 0x02
		else:
			# all 3 green
			v1, v2 = 0x20, 0x22
		if warning:
			# set the blinking flag for the leds already set
			v1 |= (v1 >> 1)
			v2 |= (v2 >> 1)
	elif charging:
		# blink all green
		v1, v2 = 0x30,0x33
	elif warning:
		# 1 red
		v1, v2 = 0x02, 0x00
	else:
		# turn off all leds
		v1, v2 = 0x11, 0x11

	write_register(device, REGISTERS.three_leds, v1, v2)


def get_notification_flags(device):
	assert device

	# Avoid a call if the device is not online,
	# or the device does not support registers.
	if device.kind is not None:
		# peripherals with protocol >= 2.0 don't support registers
		if device.protocol and device.protocol >= 2.0:
			return

	flags = read_register(device, REGISTERS.notifications)
	if flags is not None:
		assert len(flags) == 3
		return _bytes2int(flags)


def set_notification_flags(device, *flag_bits):
	assert device

	# Avoid a call if the device is not online,
	# or the device does not support registers.
	if device.kind is not None:
		# peripherals with protocol >= 2.0 don't support registers
		if device.protocol and device.protocol >= 2.0:
			return

	flag_bits = sum(int(b) for b in flag_bits)
	assert flag_bits & 0x00FFFFFF == flag_bits
	result = write_register(device, REGISTERS.notifications, _int2bytes(flag_bits, 3))
	return result is not None



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from .common import NamedInts as _NamedInts
from .hidpp10 import REGISTERS as _R, DEVICE_KIND as _DK
from .settings_templates import RegisterSettings as _RS, FeatureSettings as _FS

#
#
#

from collections import namedtuple
_DeviceDescriptor = namedtuple('_DeviceDescriptor',
				('name', 'kind', 'wpid', 'codename', 'protocol', 'registers', 'settings'))
del namedtuple

DEVICES = {}

def _D(name, codename=None, kind=None, wpid=None, protocol=None, registers=None, settings=None):
	assert name

	if kind is None:
		kind = (_DK.mouse if 'Mouse' in name
				else _DK.keyboard if 'Keyboard' in name
				else _DK.numpad if 'Number Pad' in name
				else _DK.touchpad if 'Touchpad' in name
				else _DK.trackball if 'Trackball' in name
				else None)
	assert kind is not None, 'descriptor for %s does not have kind set' % name

	# heuristic: the codename is the last word in the device name
	if codename is None and ' ' in name:
		codename = name.split(' ')[-1]
	assert codename is not None, 'descriptor for %s does not have codename set' % name

	if protocol is not None:
		# ? 2.0 devices should not have any registers
		if protocol < 2.0:
			assert settings is None or all(s._rw.kind == 1 for s in settings)
		else:
			assert registers is None
			assert settings is None or all(s._rw.kind == 2 for s in settings)

		if wpid:
			for w in wpid if isinstance(wpid, tuple) else (wpid, ):
				if protocol > 1.0:
					assert w[0:1] == '4', '%s has protocol %0.1f, wpid %s' % (name, protocol, w)
				else:
					if w[0:1] == '1':
						assert kind == _DK.mouse, '%s has protocol %0.1f, wpid %s' % (name, protocol, w)
					elif w[0:1] == '2':
						assert kind in (_DK.keyboard, _DK.numpad), '%s has protocol %0.1f, wpid %s' % (name, protocol, w)

	device_descriptor = _DeviceDescriptor(name=name, kind=kind,
					wpid=wpid, codename=codename, protocol=protocol,
					registers=registers, settings=settings)

	assert codename not in DEVICES, 'duplicate codename in device descriptors: %s' % (DEVICES[codename], )
	DEVICES[codename] = device_descriptor

	if wpid:
		if not isinstance(wpid, tuple):
			wpid = (wpid, )

		for w in wpid:
			assert w not in DEVICES, 'duplicate wpid in device descriptors: %s' % (DEVICES[w], )
			DEVICES[w] = device_descriptor

#
#
#

_PERFORMANCE_MX_DPIS = _NamedInts.range(0x81, 0x8F, lambda x: str((x - 0x80) * 100))

#
#
#

# Some HID++1.0 registers and HID++2.0 features can be discovered at run-time,
# so they are not specified here.
#
# For known registers, however, please do specify them here -- avoids
# unnecessary communication with the device and makes it easier to make certain
# decisions when querying the device's state.
#
# Specify a negative value to blacklist a certain register for a device.
#
# Usually, state registers (battery, leds, some features, etc) are only used by
# HID++ 1.0 devices, while HID++ 2.0 devices use features for the same
# functionalities. This is a rule that's been discovered by trial-and-error,
# so it may change in the future.

# Well-known registers (in hex):
#  * 00 - notification flags (all devices)
#    01 - mice: smooth scrolling
#    07 - battery status
#    09 - keyboards: FN swap (if it has the FN key)
#    0D - battery charge
#       a device may have either the 07 or 0D register available;
#       no known device uses both
#    51 - leds
#    63 - mice: DPI
#  * F1 - firmware info
# Some registers appear to be universally supported, no matter the HID++ version
# (marked with *). The rest may or may not be supported, and their values may or
# may not mean the same thing across different devices.

# The 'codename' and 'kind' fields are usually guessed from the device name,
# but in some cases (like the Logitech Cube) that heuristic fails and they have
# to be specified.
#
# The 'protocol' and 'wpid' fields are optional (they can be discovered at
# runtime), but specifying them here speeds up device discovery and reduces the
# USB traffic Solaar has to do to fully identify peripherals.
# Same goes for HID++ 2.0 feature settings (like _feature_fn_swap).
#
# The 'registers' field indicates read-only registers, specifying a state. These
# are valid (AFAIK) only to HID++ 1.0 devices.
# The 'settings' field indicates a read/write register; based on them Solaar
# generates, at runtime, the settings controls in the device panel. HID++ 1.0
# devices may only have register-based settings; HID++ 2.0 devices may only have
# feature-based settings.

# Keyboards

_D('Wireless Keyboard K230', protocol=2.0, wpid='400D')
_D('Wireless Keyboard K270', protocol=1.0,
				registers=(_R.battery_status, ),
				)
_D('Wireless Keyboard MK330')
_D('Wireless Compact Keyboard K340', protocol=1.0, wpid='2007',
				registers=(_R.battery_status, ),
				)
_D('Wireless Wave Keyboard K350', protocol=1.0, wpid='200A',
				registers=(_R.battery_status, ),
				)
_D('Wireless Keyboard K360', protocol=2.0, wpid='4004',
				settings=[
							_FS.fn_swap()
						],
				)
_D('Wireless Touch Keyboard K400', protocol=2.0, wpid=('400E', '4024'),
				settings=[
							_FS.fn_swap()
						],
				)
_D('Wireless Keyboard K520', protocol=1.0, wpid='2011',
				registers=(_R.battery_status, ),
				settings=[
							_RS.fn_swap(),
						],
				)
_D('Number Pad N545', protocol=1.0, wpid='2006',
				registers=(_R.battery_status, ),
				)
_D('Wireless Keyboard MK550')
_D('Wireless Keyboard MK700', protocol=1.0, wpid='2008',
				registers=(_R.battery_status, ),
				settings=[
							_RS.fn_swap(),
						],
				)
_D('Wireless Solar Keyboard K750', protocol=2.0, wpid='4002',
				settings=[
							_FS.fn_swap()
						],
				)
_D('Wireless Illuminated Keyboard K800', protocol=1.0, wpid='2010',
				registers=(_R.battery_status, _R.three_leds, ),
				settings=[
							_RS.fn_swap(),
							_RS.hand_detection(),
						],
				)
_D('Illuminated Living-Room Keyboard K830', protocol=2.0, wpid='4032',
				settings=[
							_FS.new_fn_swap()
						],
				)

# Mice

_D('Wireless Mouse M175')
_D('Wireless Mouse M185')
_D('Wireless Mouse M187', protocol=2.0, wpid='4019')
_D('Wireless Mouse M215', protocol=1.0, wpid='1020')
_D('Wireless Mouse M235')
_D('Wireless Mouse M305', protocol=1.0, wpid='101F',
				registers=(_R.battery_status, ),
				settings=[
							_RS.side_scroll(),
						],
				)
_D('Wireless Mouse M310', protocol=1.0, wpid='1024',
				registers=(_R.battery_status, ),
				)
_D('Wireless Mouse M315')
_D('Wireless Mouse M317')
_D('Wireless Mouse M325', protocol=2.0, wpid='400A',
				settings=[
							_FS.smooth_scroll(),
				])
_D('Wireless Mouse M345', protocol=2.0, wpid='4017')
_D('Wireless Mouse M350', protocol=1.0, wpid='101C',
				registers=(_R.battery_charge, ),
				)
_D('Wireless Mouse M505', codename='M505/B605', protocol=1.0, wpid='101D',
				registers=(_R.battery_charge, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)
_D('Wireless Mouse M510', protocol=1.0, wpid='1025',
				registers=(_R.battery_status, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)
_D('Couch Mouse M515', protocol=2.0, wpid='4007')
_D('Wireless Mouse M525', protocol=2.0, wpid='4013')
_D('Touch Mouse M600', protocol=2.0, wpid='401A')
_D('Marathon Mouse M705', protocol=1.0, wpid='101B',
				registers=(_R.battery_charge, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)
_D('Zone Touch Mouse T400')
_D('Touch Mouse T620', protocol=2.0)
_D('Logitech Cube', kind=_DK.mouse, protocol=2.0)
_D('Anywhere Mouse MX', codename='Anywhere MX', protocol=1.0, wpid='1017',
				registers=(_R.battery_charge, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)
_D('Performance Mouse MX', codename='Performance MX', protocol=1.0, wpid='101A',
				registers=(_R.battery_status, _R.three_leds, ),
				settings=[
							_RS.dpi(choices=_PERFORMANCE_MX_DPIS),
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)

_D('G7 Cordless Laser Mouse', codename='G7', protocol=1.0, wpid='1002',
				registers=(_R.battery_status, ),
				)
_D('G700 Gaming Mouse', codename='G700', protocol=1.0, wpid='1023',
				registers=(_R.battery_status, ),
				)

# Trackballs

_D('Wireless Trackball M570')

# Touchpads

_D('Wireless Rechargeable Touchpad T650', protocol=2.0, wpid='4101')
_D('Wireless Touchpad', codename='Wireless Touch', protocol=2.0, wpid='4011')

#
# Classic Nano peripherals (that don't support the Unifying protocol).
# A wpid is necessary to properly identify them.
#

_D('VX Nano Cordless Laser Mouse', codename='VX Nano', protocol=1.0, wpid=('100B', '100F'),
				registers=(_R.battery_charge, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)
_D('V450 Nano Cordless Laser Mouse', codename='V450 Nano', protocol=1.0, wpid='1011',
				registers=(_R.battery_charge, ),
				)
_D('V550 Nano Cordless Laser Mouse', codename='V550 Nano', protocol=1.0, wpid='1013',
				registers=(_R.battery_charge, ),
				settings=[
							_RS.smooth_scroll(),
							_RS.side_scroll(),
						],
				)

# Mini receiver mice

_D('MX610 Laser Cordless Mouse', codename='MX610', protocol=1.0, wpid='1001',
				registers=(_R.battery_status, ),
				)
_D('MX620 Laser Cordless Mouse', codename='MX620', protocol=1.0, wpid=('100A', '1016'),
				registers=(_R.battery_charge, ),
				)
_D('MX610 Left-Handled Mouse', codename='MX610L', protocol=1.0, wpid='1004',
				registers=(_R.battery_status, ),
				)
_D('V400 Laser Cordless Mouse', codename='V400', protocol=1.0, wpid='1003',
				registers=(_R.battery_status, ),
				)
_D('V450 Laser Cordless Mouse', codename='V450', protocol=1.0, wpid='1005',
				registers=(_R.battery_status, ),
				)
_D('VX Revolution', codename='VX Revolution', kind=_DK.mouse, protocol=1.0, wpid=('1006', '100D'),
				registers=(_R.battery_charge, ),
				)
_D('MX Air', codename='MX Air', protocol=1.0, kind=_DK.mouse, wpid=('1007', '100E'),
				registers=(_R.battery_charge, ),
				)
_D('MX Revolution', codename='MX Revolution', protocol=1.0, kind=_DK.mouse, wpid=('1008', '100C'),
				registers=(_R.battery_charge, ),
				)

# Some exotics...

_D('Fujitsu Sonic Mouse', codename='Sonic', protocol=1.0, wpid='1029')



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from gi.repository import Gtk, GLib

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger


from solaar.i18n import _
from . import icons as _icons
from logitech_receiver.status import KEYS as _K

#
#
#
_PAIRING_TIMEOUT = 30  # seconds
_STATUS_CHECK = 500  # milliseconds


def _create_page(assistant, kind, header=None, icon_name=None, text=None):
	p = Gtk.VBox(False, 8)
	assistant.append_page(p)
	assistant.set_page_type(p, kind)

	if header:
		item = Gtk.HBox(False, 16)
		p.pack_start(item, False, True, 0)

		label = Gtk.Label(header)
		label.set_alignment(0, 0)
		label.set_line_wrap(True)
		item.pack_start(label, True, True, 0)

		if icon_name:
			icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
			icon.set_alignment(1, 0)
			item.pack_start(icon, False, False, 0)

	if text:
		label = Gtk.Label(text)
		label.set_alignment(0, 0)
		label.set_line_wrap(True)
		p.pack_start(label, False, False, 0)

	p.show_all()
	return p


def _check_lock_state(assistant, receiver, count=2):
	if not assistant.is_drawable():
		if _log.isEnabledFor(_DEBUG):
			_log.debug("assistant %s destroyed, bailing out", assistant)
		return False

	if receiver.status.get(_K.ERROR):
		# receiver.status.new_device = _fake_device(receiver)
		_pairing_failed(assistant, receiver, receiver.status.pop(_K.ERROR))
		return False

	if receiver.status.new_device:
		device, receiver.status.new_device = receiver.status.new_device, None
		_pairing_succeeded(assistant, receiver, device)
		return False

	if not receiver.status.lock_open:
		if count > 0:
			# the actual device notification may arrive after the lock was paired,
			# so have a little patience
			GLib.timeout_add(_STATUS_CHECK, _check_lock_state, assistant, receiver, count - 1)
		else:
			_pairing_failed(assistant, receiver, 'failed to open pairing lock')
		return False

	return True


def _prepare(assistant, page, receiver):
	index = assistant.get_current_page()
	if _log.isEnabledFor(_DEBUG):
		_log.debug("prepare %s %d %s", assistant, index, page)

	if index == 0:
		if receiver.set_lock(False, timeout=_PAIRING_TIMEOUT):
			assert receiver.status.new_device is None
			assert receiver.status.get(_K.ERROR) is None
			spinner = page.get_children()[-1]
			spinner.start()
			GLib.timeout_add(_STATUS_CHECK, _check_lock_state, assistant, receiver)
			assistant.set_page_complete(page, True)
		else:
			GLib.idle_add(_pairing_failed, assistant, receiver, 'the pairing lock did not open')
	else:
		assistant.remove_page(0)


def _finish(assistant, receiver):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("finish %s", assistant)
	assistant.destroy()
	receiver.status.new_device = None
	if receiver.status.lock_open:
		receiver.set_lock()
	else:
		receiver.status[_K.ERROR] = None


def _pairing_failed(assistant, receiver, error):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("%s fail: %s", receiver, error)

	assistant.commit()

	header = _("Pairing failed") + ': ' + _(str(error)) + '.'
	if 'timeout' in str(error):
		text = _("Make sure your device is within range, and has a decent battery charge.")
	elif str(error) == 'device not supported':
		text = _("A new device was detected, but it is not compatible with this receiver.")
	elif 'many' in str(error):
		text = _("The receiver only supports %d paired device(s).")
	else:
		text = _("No further details are available about the error.")
	_create_page(assistant, Gtk.AssistantPageType.SUMMARY, header, 'dialog-error', text)

	assistant.next_page()
	assistant.commit()


def _pairing_succeeded(assistant, receiver, device):
	assert device
	if _log.isEnabledFor(_DEBUG):
		_log.debug("%s success: %s", receiver, device)

	page = _create_page(assistant, Gtk.AssistantPageType.SUMMARY)

	header = Gtk.Label(_("Found a new device") + ':')
	header.set_alignment(0.5, 0)
	page.pack_start(header, False, False, 0)

	device_icon = Gtk.Image()
	icon_set = _icons.device_icon_set(device.name, device.kind)
	device_icon.set_from_icon_set(icon_set, Gtk.IconSize.LARGE)
	device_icon.set_alignment(0.5, 1)
	page.pack_start(device_icon, True, True, 0)

	device_label = Gtk.Label()
	device_label.set_markup('<b>%s</b>' % device.name)
	device_label.set_alignment(0.5, 0)
	page.pack_start(device_label, True, True, 0)

	hbox = Gtk.HBox(False, 8)
	hbox.pack_start(Gtk.Label(' '), False, False, 0)
	hbox.set_property('expand', False)
	hbox.set_property('halign', Gtk.Align.CENTER)
	page.pack_start(hbox, False, False, 0)

	def _check_encrypted(dev):
		if assistant.is_drawable():
			if device.status.get(_K.LINK_ENCRYPTED) == False:
				hbox.pack_start(Gtk.Image.new_from_icon_name('security-low', Gtk.IconSize.MENU), False, False, 0)
				hbox.pack_start(Gtk.Label(_("The wireless link is not encrypted") + '!'), False, False, 0)
				hbox.show_all()
			else:
				return True
	GLib.timeout_add(_STATUS_CHECK, _check_encrypted, device)

	page.show_all()

	assistant.next_page()
	assistant.commit()


def create(receiver):
	assert receiver is not None
	assert receiver.kind is None

	assistant = Gtk.Assistant()
	assistant.set_title(receiver.name + ': ' + _("pair new device"))
	assistant.set_icon_name('list-add')

	assistant.set_size_request(400, 240)
	assistant.set_resizable(False)
	assistant.set_role('pair-device')

	page_intro = _create_page(assistant, Gtk.AssistantPageType.PROGRESS,
					_("Turn on the device you want to pair."), 'preferences-desktop-peripherals',
					_("If the device is already turned on,\nturn if off and on again."))
	spinner = Gtk.Spinner()
	spinner.set_visible(True)
	page_intro.pack_end(spinner, True, True, 24)

	assistant.connect('prepare', _prepare, receiver)
	assistant.connect('cancel', _finish, receiver)
	assistant.connect('close', _finish, receiver)

	return assistant



#!/usr/bin/env python

from glob import glob as _glob
from distutils.core import setup

autostart_path = '/etc/xdg/autostart'

import sys
backup_path_0 = sys.path[0]
sys.path[0] = backup_path_0 + '/lib'
from solaar import NAME, __version__
sys.path[0] = backup_path_0

if 'install' in sys.argv:
	# naively guess where the autostart .desktop file should be installed
	if '--prefix' in sys.argv or '--home' in sys.argv:
		autostart_path = 'etc/xdg/autostart'
	elif '--user' in sys.argv:
		from os import environ
		from os import path
		xdg_config_home = environ.get('XDG_CONFIG_HOME', path.expanduser(path.join('~', '.config')))
		autostart_path = path.join(xdg_config_home, 'autostart')
		del environ, path, xdg_config_home

del sys, backup_path_0


def _data_files():
	from os.path import dirname as _dirname

	yield 'share/solaar/icons', _glob('share/solaar/icons/solaar*.svg')
	yield 'share/solaar/icons', _glob('share/solaar/icons/light_*.png')
	yield 'share/icons/hicolor/scalable/apps', ['share/solaar/icons/solaar.svg']

	for mo in _glob('share/locale/*/LC_MESSAGES/solaar.mo'):
		yield _dirname(mo), [mo]

	yield 'share/applications', ['share/applications/solaar.desktop']
	yield autostart_path, ['share/applications/solaar.desktop']

	del _dirname


setup(name=NAME.lower(),
		version=__version__,
		description='Linux devices manager for the Logitech Unifying Receiver.',
		long_description='''
Solaar is a Linux device manager for Logitech's Unifying Receiver peripherals.
It is able to pair/unpair devices to the receiver, and for some devices read
battery status.
'''.strip(),
		author='Daniel Pavel',
		author_email='daniel.pavel@gmail.com',
		license='GPLv2',
		url='http://pwr.github.io/Solaar/',
		classifiers=[
			'Development Status :: 4 - Beta',
			'Environment :: X11 Applications :: GTK',
			'Environment :: Console',
			'Intended Audience :: End Users/Desktop',
			'License :: DFSG approved',
			'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
			'Natural Language :: English',
			'Programming Language :: Python :: 2.7',
			'Programming Language :: Python :: 3.2',
			'Operating System :: POSIX :: Linux',
			'Topic :: Utilities',
			],

		platforms=['linux'],
		requires=['pyudev (>= 0.13)', 'gi.repository.GObject (>= 2.0)', 'gi.repository.Gtk (>= 3.0)'],

		package_dir={'': 'lib'},
		packages=['hidapi', 'logitech_receiver', 'solaar', 'solaar.ui', 'solaar.cli'],
		data_files=list(_data_files()),
		scripts=_glob('bin/*'),
	)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


import argparse as _argparse
import sys as _sys

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger


from solaar import NAME

#
#
#

def _create_parser():
	parser = _argparse.ArgumentParser(prog=NAME.lower(), add_help=False,
					epilog='For details on individual actions, run `%s <action> --help`.' % NAME.lower())
	subparsers = parser.add_subparsers(title='actions',
					help='optional action to perform')

	sp = subparsers.add_parser('show', help='show information about devices')
	sp.add_argument('device', nargs='?', default='all',
					help='device to show information about; may be a device number (1..6), a serial, '
						'a substring of a device\'s name, or "all" (the default)')
	sp.set_defaults(action='show')

	sp = subparsers.add_parser('config', help='read/write device-specific settings',
								epilog='Please note that configuration only works on active devices.')
	sp.add_argument('device',
					help='device to configure; may be a device number (1..6), a device serial, '
							'or at least 3 characters of a device\'s name')
	sp.add_argument('setting', nargs='?',
					help='device-specific setting; leave empty to list available settings')
	sp.add_argument('value', nargs='?',
					help='new value for the setting')
	sp.set_defaults(action='config')

	sp = subparsers.add_parser('pair', help='pair a new device',
					epilog='The Logitech Unifying Receiver supports up to 6 paired devices at the same time.')
	sp.add_argument('receiver', nargs='?',
					help='select a certain receiver when more than one is present')
	sp.set_defaults(action='pair')

	sp = subparsers.add_parser('unpair', help='unpair a device')
	sp.add_argument('device',
					help='device to unpair; may be a device number (1..6), a serial, '
						'or a substring of a device\'s name.')
	sp.set_defaults(action='unpair')

	return parser, subparsers.choices


_cli_parser, actions = _create_parser()
print_help = _cli_parser.print_help


def _receivers(dev_path=None):
	from logitech_receiver import Receiver
	from logitech_receiver.base import receivers
	for dev_info in receivers():
		if dev_path is not None and dev_path != dev_info.path:
			continue
		try:
			r = Receiver.open(dev_info)
			if _log.isEnabledFor(_DEBUG):
				_log.debug("[%s] => %s", dev_info.path, r)
			if r:
				yield r
		except Exception as e:
			_log.exception('opening ' + str(dev_info))
			_sys.exit("%s: error: %s" % (NAME, str(e)))


def _find_receiver(receivers, name):
	assert receivers
	assert name

	for r in receivers:
		if name in r.name.lower() or name == r.serial.lower():
			return r


def _find_device(receivers, name):
	assert receivers
	assert name

	number = None
	if len(name) == 1:
		try:
			number = int(name)
		except:
			pass
		else:
			assert not (number < 0)
			if number > 6: number = None

	for r in receivers:
		if number and number <= r.max_devices:
			dev = r[number]
			if dev:
				return dev

		for dev in r:
			if (name == dev.serial.lower() or
				name == dev.codename.lower() or
				name == str(dev.kind).lower() or
				name in dev.name.lower()):
				return dev

	raise Exception("no device found matching '%s'" % name)


def run(cli_args=None, hidraw_path=None):
	if cli_args:
		action = cli_args[0]
		args = _cli_parser.parse_args(cli_args)
	else:
		args = _cli_parser.parse_args()
		# Python 3 has an undocumented 'feature' that breaks parsing empty args
		# http://bugs.python.org/issue16308
		if not 'cmd' in args:
			_cli_parser.print_usage(_sys.stderr)
			_sys.stderr.write('%s: error: too few arguments\n' % NAME.lower())
			_sys.exit(2)
		action = args.action
	assert action in actions

	try:
		c = list(_receivers(hidraw_path))
		if not c:
			raise Exception('Logitech receiver not found')

		from importlib import import_module
		m = import_module('.' + action, package=__name__)
		m.run(c, args, _find_receiver, _find_device)
	except Exception as e:
		_sys.exit('%s: error: %s' % (NAME.lower(), e))



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Reprogrammable keys information

from __future__ import absolute_import, division, print_function, unicode_literals


from .common import NamedInts as _NamedInts

# <controls.xml awk -F\" '/<Control /{sub(/^LD_FINFO_(CTRLID_)?/, "", $2);printf("\t%s=0x%04X,\n", $2, $4)}' | sort -t= -k2
CONTROL = _NamedInts(
	Volume_Up=0x0001,
	Volume_Down=0x0002,
	Mute=0x0003,
	Play__Pause=0x0004,
	Next=0x0005,
	Previous=0x0006,
	Stop=0x0007,
	Application_Switcher=0x0008,
	BURN=0x0009,
	Calculator=0x000A,
	CALENDAR=0x000B,
	CLOSE=0x000C,
	EJECT=0x000D,
	Mail=0x000E,
	HELP_AS_HID=0x000F,
	HELP_AS_F1=0x0010,
	LAUNCH_WORD_PROC=0x0011,
	LAUNCH_SPREADSHEET=0x0012,
	LAUNCH_PRESENTATION=0x0013,
	UNDO_AS_CTRL_Z=0x0014,
	UNDO_AS_HID=0x0015,
	REDO_AS_CTRL_Y=0x0016,
	REDO_AS_HID=0x0017,
	PRINT_AS_CTRL_P=0x0018,
	PRINT_AS_HID=0x0019,
	SAVE_AS_CTRL_S=0x001A,
	SAVE_AS_HID=0x001B,
	PRESET_A=0x001C,
	PRESET_B=0x001D,
	PRESET_C=0x001E,
	PRESET_D=0x001F,
	FAVORITES=0x0020,
	GADGETS=0x0021,
	MY_HOME=0x0022,
	GADGETS_AS_WIN_G=0x0023,
	MAXIMIZE_AS_HID=0x0024,
	MAXIMIZE_AS_WIN_SHIFT_M=0x0025,
	MINIMIZE_AS_HID=0x0026,
	MINIMIZE_AS_WIN_M=0x0027,
	MEDIA_PLAYER=0x0028,
	MEDIA_CENTER_LOGI=0x0029,
	MEDIA_CENTER_MSFT=0x002A, # Should not be used as it is not reprogrammable under Windows
	CUSTOM_MENU=0x002B,
	MESSENGER=0x002C,
	MY_DOCUMENTS=0x002D,
	MY_MUSIC=0x002E,
	WEBCAM=0x002F,
	MY_PICTURES=0x0030,
	MY_VIDEOS=0x0031,
	MY_COMPUTER_AS_HID=0x0032,
	MY_COMPUTER_AS_WIN_E=0x0033,
	LAUNC_PICTURE_VIEWER=0x0035,
	ONE_TOUCH_SEARCH=0x0036,
	PRESET_1=0x0037,
	PRESET_2=0x0038,
	PRESET_3=0x0039,
	PRESET_4=0x003A,
	RECORD=0x003B,
	INTERNET_REFRESH=0x003C,
	ROTATE_RIGHT=0x003D,
	SEARCH=0x003E,
	SHUFFLE=0x003F,
	SLEEP=0x0040,
	INTERNET_STOP=0x0041,
	SYNCHRONIZE=0x0042,
	ZOOM=0x0043,
	ZOOM_IN_AS_HID=0x0044,
	ZOOM_IN_AS_CTRL_WHEEL=0x0045,
	ZOOM_IN_AS_CLTR_PLUS=0x0046,
	ZOOM_OUT_AS_HID=0x0047,
	ZOOM_OUT_AS_CTRL_WHEEL=0x0048,
	ZOOM_OUT_AS_CLTR_MINUS=0x0049,
	ZOOM_RESET=0x004A,
	ZOOM_FULL_SCREEN=0x004B,
	PRINT_SCREEN=0x004C,
	PAUSE_BREAK=0x004D,
	SCROLL_LOCK=0x004E,
	CONTEXTUAL_MENU=0x004F,
	LEFT_CLICK=0x0050,
	RIGHT_CLICK=0x0051,
	MIDDLE_BUTTON=0x0052,
	BACK_AS_BUTTON_4=0x0053,
	BACK_AS_HID=0x0054,
	BACK_AS_ALT_WIN_ARROW=0x0055,
	FORWARD_AS_BUTTON_5=0x0056,
	FORWARD_AS_HID=0x0057,
	FORWARD_AS_ALT_WIN_ARROW=0x0058,
	BUTTON_6=0x0059,
	LEFT_SCROLL_AS_BUTTON_7=0x005A,
	LEFT_SCROLL_AS_AC_PAN=0x005B,
	RIGHT_SCROLL_AS_BUTTON_8=0x005C,
	RIGHT_SCROLL_AS_AC_PAN=0x005D,
	BUTTON_9=0x005E,
	BUTTON_10=0x005F,
	BUTTON_11=0x0060,
	BUTTON_12=0x0061,
	BUTTON_13=0x0062,
	BUTTON_14=0x0063,
	BUTTON_15=0x0064,
	BUTTON_16=0x0065,
	BUTTON_17=0x0066,
	BUTTON_18=0x0067,
	BUTTON_19=0x0068,
	BUTTON_20=0x0069,
	BUTTON_21=0x006A,
	BUTTON_22=0x006B,
	BUTTON_23=0x006C,
	BUTTON_24=0x006D,
	SHOW_DESKTOP=0x006E,
	Lock_PC=0x006F,
	FN_F1=0x0070,
	FN_F2=0x0071,
	FN_F3=0x0072,
	FN_F4=0x0073,
	FN_F5=0x0074,
	FN_F6=0x0075,
	FN_F7=0x0076,
	FN_F8=0x0077,
	FN_F9=0x0078,
	FN_F10=0x0079,
	FN_F11=0x007A,
	FN_F12=0x007B,
	FN_F13=0x007C,
	FN_F14=0x007D,
	FN_F15=0x007E,
	FN_F16=0x007F,
	FN_F17=0x0080,
	FN_F18=0x0081,
	FN_F19=0x0082,
	IOS_HOME=0x0083,
	ANDROID_HOME=0x0084,
	ANDROID_MENU=0x0085,
	ANDROID_SEARCH=0x0086,
	ANDROID_BACK=0x0087,
	HOME_COMBO=0x0088,
	LOCK_COMBO=0x0089,
	IOS_VIRTUAL_KEYBOARD=0x008A,
	IOS_LANGUAGE_SWICH=0x008B,
	MAC_EXPOSE=0x008C,
	MAC_DASHBOARD=0x008D,
	WIN7_SNAP_LEFT=0x008E,
	WIN7_SNAP_RIGHT=0x008F,
	WIN7_MINIMIZE_AS_WIN_ARROW=0x0090,
	WIN7_MAXIMIZE_AS_WIN_ARROW=0x0091,
	WIN7_STRETCH_UP=0x0092,
	WIN7_MONITOR_SWITCH_AS_WIN_SHIFT_LEFTARROW=0x0093,
	WIN7_MONITOR_SWITCH_AS_WIN_SHIFT_RIGHTARROW=0x0094,
	WIN7_SHOW_PRESENTATION_MODE=0x0095,
	WIN7_SHOW_MOBILITY_CENTER=0x0096,
	ANALOG_HSCROLL=0x0097,
	METRO_APPSWITCH=0x009F,
	METRO_APPBAR=0x00A0,
	METRO_CHARMS=0x00A1,
	CALC_VKEYBOARD=0x00A2,
	METRO_SEARCH=0x00A3,
	COMBO_SLEEP=0x00A4,
	METRO_SHARE=0x00A5,
	METRO_SETTINGS=0x00A6,
	METRO_DEVICES=0x00A7,
	METRO_START_SCREEN=0x00A9,
	ZOOMIN=0x00AA,
	ZOOMOUT=0x00AB,
	BACK_HSCROLL=0x00AC,
	SHOW_DESKTOP_HPP=0x00AE,
)
CONTROL._fallback = lambda x: 'unknown:%04X' % x

# <tasks.xml awk -F\" '/<Task /{gsub(/ /, "_", $6); printf("\t%s=0x%04X,\n", $6, $4)}'
TASK = _NamedInts(
	Volume_Up=0x0001,
	Volume_Down=0x0002,
	Mute=0x0003,
	# Multimedia tasks:
	Play__Pause=0x0004,
	Next=0x0005,
	Previous=0x0006,
	Stop=0x0007,

	Application_Switcher=0x0008,
	BurnMediaPlayer=0x0009,
	Calculator=0x000A,
	Calendar=0x000B,
	Close_Application=0x000C,
	Eject=0x000D,
	Email=0x000E,
	Help=0x000F,
	OffDocument=0x0010,
	OffSpreadsheet=0x0011,
	OffPowerpnt=0x0012,
	Undo=0x0013,
	Redo=0x0014,
	Print=0x0015,
	Save=0x0016,
	SmartKeySet=0x0017,
	Favorites=0x0018,
	GadgetsSet=0x0019,
	HomePage=0x001A,
	WindowsRestore=0x001B,
	WindowsMinimize=0x001C,
	Music=0x001D, # also known as MediaPlayer

	# Both 0x001E and 0x001F are known as MediaCenterSet
	Media_Center_Logitech=0x001E,
	Media_Center_Microsoft=0x001F,

	UserMenu=0x0020,
	Messenger=0x0021,
	PersonalFolders=0x0022,
	MyMusic=0x0023,
	Webcam=0x0024,
	PicturesFolder=0x0025,
	MyVideos=0x0026,
	My_Computer=0x0027,
	PictureAppSet=0x0028,
	Search=0x0029, # also known as AdvSmartSearch
	RecordMediaPlayer=0x002A,
	BrowserRefresh=0x002B,
	RotateRight=0x002C,
	SearchForFiles=0x002D,
	MM_SHUFFLE=0x002E,
	Sleep=0x002F, # also known as StandBySet
	BrowserStop=0x0030,
	OneTouchSync=0x0031,
	ZoomSet=0x0032,
	ZoomBtnInSet2=0x0033,
	ZoomBtnInSet=0x0034,
	ZoomBtnOutSet2=0x0035,
	ZoomBtnOutSet=0x0036,
	ZoomBtnResetSet=0x0037,
	LeftClick=0x0038,
	RightClick=0x0039,
	MiddleMouseButton=0x003A,
	Back=0x003B,
	BackEx=0x003C,
	BrowserForward=0x003D,
	BrowserForwardEx=0x003E,
	HorzScrollLeftSet=0x003F,
	HorzScrollRightSet=0x0040,
	QuickSwitch=0x0041,
	BatteryStatus=0x0042,
	ShowDesktop=0x0043,
	WindowsLock=0x0044,
	FileLauncher=0x0045,
	FolderLauncher=0x0046,
	GotoWebAddress=0x0047,
	GenericMouseButton=0x0048,
	KeystrokeAssignment=0x0049,
	LaunchProgram=0x004A,
	MinMaxWindow=0x004B,
	VOLUMEMUTE_NoOSD=0x004C,
	New=0x004D,
	Copy=0x004E,
	CruiseDown=0x004F,
	CruiseUp=0x0050,
	Cut=0x0051,
	Do_Nothing=0x0052,
	PageDown=0x0053,
	PageUp=0x0054,
	Paste=0x0055,
	SearchPicture=0x0056,
	Reply=0x0057,
	PhotoGallerySet=0x0058,
	MM_REWIND=0x0059,
	MM_FASTFORWARD=0x005A,
	Send=0x005B,
	ControlPanel=0x005C,
	UniversalScroll=0x005D,
	AutoScroll=0x005E,
	GenericButton=0x005F,
	MM_NEXT=0x0060,
	MM_PREVIOUS=0x0061,
	Do_Nothing_One=0x0062, # also known as Do_Nothing
	SnapLeft=0x0063,
	SnapRight=0x0064,
	WinMinRestore=0x0065,
	WinMaxRestore=0x0066,
	WinStretch=0x0067,
	SwitchMonitorLeft=0x0068,
	SwitchMonitorRight=0x0069,
	ShowPresentation=0x006A,
	ShowMobilityCenter=0x006B,
	HorzScrollNoRepeatSet=0x006C,
	TouchBackForwardHorzScroll=0x0077,
	MetroAppSwitch=0x0078,
	MetroAppBar=0x0079,
	MetroCharms=0x007A,
	Calculator_VKEY=0x007B, # also known as Calculator
	MetroSearch=0x007C,
	MetroStartScreen=0x0080,
	MetroShare=0x007D,
	MetroSettings=0x007E,
	MetroDevices=0x007F,
	MetroBackLeftHorz=0x0082,
	MetroForwRightHorz=0x0083,
	Win8_Back=0x0084, # also known as MetroCharms
	Win8_Forward=0x0085, # also known as AppSwitchBar
	Win8Charm_Appswitch_GifAnimation=0x0086,
	Win8BackHorzLeft=0x008B, # also known as Back
	Win8ForwardHorzRight=0x008C, # also known as BrowserForward
	MetroSearch2=0x0087,
	MetroShare2=0x0088,
	MetroSettings2=0x008A,
	MetroDevices2=0x0089,
	Win8MetroWin7Forward=0x008D, # also known as MetroStartScreen
	Win8ShowDesktopWin7Back=0x008E, # also known as ShowDesktop
	MetroApplicationSwitch=0x0090, # also known as MetroStartScreen
	ShowUI=0x0092,
)
TASK._fallback = lambda x: 'unknown:%04X' % x

KEY_FLAG = _NamedInts(
	reprogrammable=0x10,
	FN_sensitive=0x08,
	nonstandard=0x04,
	is_FN=0x02,
	mse=0x01
)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import threading as _threading
# from time import time as _timestamp

# for both Python 2 and 3
try:
	from Queue import Queue as _Queue
except ImportError:
	from queue import Queue as _Queue

from logging import getLogger, DEBUG as _DEBUG, INFO as _INFO
_log = getLogger(__name__)
del getLogger


from . import base as _base

#
#
#

class _ThreadedHandle(object):
	"""A thread-local wrapper with different open handles for each thread.

	Closing a ThreadedHandle will close all handles.
	"""

	__slots__ = ('path', '_local', '_handles', '_listener')

	def __init__(self, listener, path, handle):
		assert listener is not None
		assert path is not None
		assert handle is not None
		assert isinstance(handle, int)

		self._listener = listener
		self.path = path
		self._local = _threading.local()
		# take over the current handle for the thread doing the replacement
		self._local.handle = handle
		self._handles = [handle]

	def _open(self):
		handle = _base.open_path(self.path)
		if handle is None:
			_log.error("%r failed to open new handle", self)
		else:
			# if _log.isEnabledFor(_DEBUG):
			# 	_log.debug("%r opened new handle %d", self, handle)
			self._local.handle = handle
			self._handles.append(handle)
			return handle

	def close(self):
		if self._local:
			self._local = None
			handles, self._handles = self._handles, []
			if _log.isEnabledFor(_DEBUG):
				_log.debug("%r closing %s", self, handles)
			for h in handles:
				_base.close(h)

	@property
	def notifications_hook(self):
		if self._listener:
			assert isinstance(self._listener, _threading.Thread)
			if _threading.current_thread() == self._listener:
				return self._listener._notifications_hook

	def __del__(self):
		self._listener = None
		self.close()

	def __index__(self):
		if self._local:
			try:
				return self._local.handle
			except:
				return self._open()
	__int__ = __index__

	def __str__(self):
		if self._local:
			return str(int(self))
	__unicode__ = __str__

	def __repr__(self):
		return '<_ThreadedHandle(%s)>' % self.path

	def __bool__(self):
		return bool(self._local)
	__nonzero__ = __bool__

#
#
#

# How long to wait during a read for the next packet, in seconds
# Ideally this should be rather long (10s ?), but the read is blocking
# and this means that when the thread is signalled to stop, it would take
# a while for it to acknowledge it.
# Forcibly closing the file handle on another thread does _not_ interrupt the
# read on Linux systems.
_EVENT_READ_TIMEOUT = 0.4  # in seconds

# After this many reads that did not produce a packet, call the tick() method.
# This only happens if tick_period is enabled (>0) for the Listener instance.
# _IDLE_READS = 1 + int(5 // _EVENT_READ_TIMEOUT)  # wait at least 5 seconds between ticks


class EventsListener(_threading.Thread):
	"""Listener thread for notifications from the Unifying Receiver.

	Incoming packets will be passed to the callback function in sequence.
	"""
	def __init__(self, receiver, notifications_callback):
		super(EventsListener, self).__init__(name=self.__class__.__name__ + ':' + receiver.path.split('/')[2])

		self.daemon = True
		self._active = False

		self.receiver = receiver
		self._queued_notifications = _Queue(16)
		self._notifications_callback = notifications_callback

		# self.tick_period = 0

	def run(self):
		self._active = True

		# replace the handle with a threaded one
		self.receiver.handle = _ThreadedHandle(self, self.receiver.path, self.receiver.handle)
		# get the right low-level handle for this thead
		ihandle = int(self.receiver.handle)
		if _log.isEnabledFor(_INFO):
			_log.info("started with %s (%d)", self.receiver, ihandle)

		self.has_started()

		# last_tick = 0
		# the first idle read -- delay it a bit, and make sure to stagger
		# idle reads for multiple receivers
		# idle_reads = _IDLE_READS + (ihandle % 5) * 2

		while self._active:
			if self._queued_notifications.empty():
				try:
					# _log.debug("read next notification")
					n = _base.read(ihandle, _EVENT_READ_TIMEOUT)
				except _base.NoReceiver:
					_log.warning("receiver disconnected")
					self.receiver.close()
					break

				if n:
					n = _base.make_notification(*n)
			else:
				# deliver any queued notifications
				n = self._queued_notifications.get()

			if n:
				# if _log.isEnabledFor(_DEBUG):
				# 	_log.debug("%s: processing %s", self.receiver, n)
				try:
					self._notifications_callback(n)
				except:
					_log.exception("processing %s", n)

			# elif self.tick_period:
			# 	idle_reads -= 1
			# 	if idle_reads <= 0:
			# 		idle_reads = _IDLE_READS
			# 		now = _timestamp()
			# 		if now - last_tick >= self.tick_period:
			# 			last_tick = now
			# 			self.tick(now)

		del self._queued_notifications
		self.has_stopped()

	def stop(self):
		"""Tells the listener to stop as soon as possible."""
		self._active = False

	def has_started(self):
		"""Called right after the thread has started, and before it starts
		reading notification packets."""
		pass

	def has_stopped(self):
		"""Called right before the thread stops."""
		pass

	# def tick(self, timestamp):
	# 	"""Called about every tick_period seconds."""
	# 	pass

	def _notifications_hook(self, n):
		# Only consider unhandled notifications that were sent from this thread,
		# i.e. triggered by a callback handling a previous notification.
		assert _threading.current_thread() == self
		if self._active:  # and _threading.current_thread() == self:
			# if _log.isEnabledFor(_DEBUG):
			# 	_log.debug("queueing unhandled %s", n)
			self._queued_notifications.put(n)

	def __bool__(self):
		return bool(self._active and self.receiver)
	__nonzero__ = __bool__
